from typing import Union

from httpx.config import (TimeoutTypes, UNSET, UnsetType)

from pysjtu import consts
from pysjtu import models
from pysjtu.client.base import BaseClient


class ScheduleMixin(BaseClient):
    def __init__(self):
        super().__init__()

    def schedule(self, year: int, term: int, timeout: Union[TimeoutTypes, UnsetType] = UNSET) \
            -> models.Results[models.ScheduleCourse]:
        """
        Fetch your course schedule of specific year & term.

        :param year: year for the new :class:`Schedule` object.
        :param term: term for the new :class:`Schedule` object.
        :param timeout: (optional) How long to wait for the server to send data before giving up.
        :return: A new :class:`Schedule` object.
        :rtype: :class:`Schedule`
        """
        raw = self._session.post(consts.SCHEDULE_URL, data={"xnm": year, "xqm": consts.TERMS[term]}, timeout=timeout)
        schedule = models.Schedule(year, term)
        schedule.load(raw.json()["kbList"])  # type: ignore
        return schedule



from __future__ import annotations

from dataclasses import dataclass
from typing import Callable, List, Optional, Tuple, Union

from pysjtu.models.base import Gender, LazyResult, PARTIAL, Result
from pysjtu.utils import elfhash


@dataclass
class LessonTime:
    weekday: int
    week: List[Union[range, int]]
    time: List[range]


# noinspection PyAbstractClass
@dataclass
class SelectionSharedInfo:
    """
    A model which contains shared information in this round of selection.

    :param term: current term when querying.
    :type term: str
    :param selection_year: year of selected courses.
    :type selection_year: int
    :param selection_term: term of selected courses.
    :type selection_term: int
    :param major_id: internal major id.
    :type major_id: str
    :param student_grade: year of enrollment.
    :type student_grade: str
    :param natural_class_id: class id of the administrative class of the student.
    :type natural_class_id: str
    :param self_selecting_status: unknown parameter.
    :type self_selecting_status: int
    :param ccdm: unknown parameter.
    :type ccdm: str
    :param student_type_code: unknown parameter.
    :type student_type_code: int
    :param gender: student's gender.
    :type gender: Gender
    :param field_id: student's professional field.
    :type field_id: str
    :param student_background: unknown parameter.
    :type student_background: int
    """
    term: str
    selection_year: int
    selection_term: int
    major_id: str
    student_grade: int
    natural_class_id: str
    self_selecting_status: int
    ccdm: str
    student_type_code: int
    gender: Gender
    field_id: str
    student_background: int


# noinspection PyAbstractClass
@dataclass
class SelectionSector(Result):
    """
    A model which describes a course sector in this round of selection.

    :param name: sector name.
    :type name: str
    :param shared_info: shared information in this round of selection.
    :type shared_info: SelectionSharedInfo
    :param task_type: unknown parameter.
    :type task_type: int
    :param xkly: unknown parameter.
    :type xkly: int
    :param pe_op_param: unknown parameter. (translation: the parameters updated in operations related to PE lessons.)
    :type pe_op_param: int
    :param sector_type_id: unknown parameter.
    :type sector_type_id: str
    :param include_other_grades: include courses from other grades.
    :type include_other_grades: bool
    :param include_other_majors: include courses from other majors.
    :type include_other_majors: bool
    :param sfznkx: unknown parameter.
    :type sfznkx: str
    :param zdkxms: unknown parameter.
    :type zdkxms: int
    :param txbsfrl: unknown parameter. (used when deregistering courses.)
    :type txbsfrl: int
    :param kkbk: unknown parameter.
    :type kkbk: int
    :param course_type_code: unknown parameter.
    :type course_type_code: str
    :param xkkz_id: unknown parameter.
    :type xkkz_id: str
    """
    task_type: int
    xkly: int
    pe_op_param: int
    sector_type_id: str
    include_other_grades: bool
    include_other_majors: bool
    sfznkx: bool
    zdkxms: int
    txbsfrl: int
    kkbk: int
    course_type_code: Optional[str] = None
    name: Optional[str] = None
    xkkz_id: Optional[str] = None
    shared_info: Optional[SelectionSharedInfo] = None
    _func_classes: Callable = None
    _hash: Optional[int] = None

    def __hash__(self):
        if not self._hash:
            self._hash = elfhash(self.name)
        return self._hash

    def __repr__(self):
        return f"<SelectionSector {self.name}>"

    @property
    def classes(self) -> List[SelectionClass]:
        """
        Selectable classes in this course sector.
        """
        return self._func_classes()


@dataclass
class SelectionClass(LazyResult):
    """
    A model which describes a selectable class in this round of selection.

    :param name: literal name of the course.
    :type name: str
    :param credit: credits that the course provides.
    :type credit: int
    :param course_id: course id.
    :type course_id: str
    :param course_id: internal course id.
    :type course_id: str
    :param class_name: class name (constant between years).
    :type class_name: str
    :param class_id: class id (variable between years).
    :type class_id: str
    :param students_elected: number of students registered for this course.
    :type students_elected: int
    :param students_planned: number of students planned when setting this course.
    :type students_planned: int
    :param register_id: dynamic id used when (de)registering for this class.
    :param teacher: the teachers who offer this course.
    :type teacher: List[Tuple[str]]
    :param locations: the places where classes are given.
    :type locations: List[str]
    :param time: the time when the class is given.
    :type time: LessonTime
    :param course_type: course type. (eg. general, required, optional, ...)
    :type course_type: str
    :param remark: remarks of this class.
    :type remark: str
    :param sector: the sector which this course lies in.
    """
    name: str
    credit: float
    course_id: str
    internal_course_id: str
    class_name: str
    class_id: str
    students_registered: int
    register_id: str = PARTIAL
    teachers: List[Tuple[str]] = PARTIAL
    locations: List[str] = PARTIAL
    time: LessonTime = PARTIAL
    course_type: List[str] = PARTIAL
    remark: Optional[str] = PARTIAL
    students_planned: int = PARTIAL
    sector: SelectionSector = None
    _load_func: Callable = None

    def __repr__(self):
        return f"<SelectionClass {self.class_name} {self.name}>"

    def is_registered(self, timeout=10) -> bool:
        """
        Check whether the student has registered for this class.

        :param timeout: (optional) How long to wait for the server to send data before giving up.
        :return: A boolean value indicates the registration status.
        :rtype: bool
        """
        raise NotImplementedError  # pragma: no cover

    def register(self, timeout=10):
        """
        Register for this class.

        :param timeout: (optional) How long to wait for the server to send data before giving up.
        :return: None
        :raises: :exc:`pysjtu.exceptions.RegistrationException`
        :raises: :exc:`pysjtu.exceptions.FullCapacityException`
        :raises: :exc:`pysjtu.exceptions.TimeConflictException`
        :raises: :exc:`pysjtu.exceptions.SelectionNotAvailableException`
        """
        raise NotImplementedError  # pragma: no cover

    def deregister(self, timeout=10):
        """
        Drop this class.

        :param timeout: (optional) How long to wait for the server to send data before giving up.
        :return: None
        :raises: :exc:`pysjtu.exceptions.DeregistrationException`
        :raises: :exc:`pysjtu.exceptions.SelectionNotAvailableException`
        """
        raise NotImplementedError  # pragma: no cover



from functools import lru_cache, partial
from typing import List

from pysjtu import consts
from pysjtu.client.base import BaseClient
from pysjtu.exceptions import DeregistrationException, FullCapacityException, RegistrationException, \
    SelectionClassFetchException, SelectionNotAvailableException, TimeConflictException
from pysjtu.models.selection import SelectionClass, SelectionSector, SelectionSharedInfo
from pysjtu.parser.selection import parse_sector, parse_sectors, parse_shared_info
from pysjtu.schemas.selection import SelectionClassSchema, SelectionCourseSchema, SelectionSectorSchema, \
    SelectionSharedInfoSchema


class SelectionMixin(BaseClient):
    def __init__(self):
        super().__init__()
        self._fetch_selection_classes = lru_cache(maxsize=1024)(self._fetch_selection_classes)
        self._get_selection_classes = lru_cache(maxsize=16)(self._get_selection_classes)

    def _class_is_registered(self, _class: SelectionClass, timeout=10) -> bool:
        payload = {
            "jxb_id": _class.register_id,
            "xkkz_id": _class.sector.xkkz_id,
            "xnm": _class.sector.shared_info.selection_year,
            "xqm": _class.sector.shared_info.selection_term
        }
        is_registered = self._session.post(f"{consts.SELECTION_IS_REGISTERED}{self.student_id}", data=payload,
                                           timeout=timeout).json()
        return is_registered == "1"

    def _class_register(self, _class: SelectionClass, timeout=10):
        payload = {
            "jxb_ids": _class.register_id,
            "kch_id": _class.internal_course_id,
            "qz": 0
        }
        register = self._session.post(f"{consts.SELECTION_REGISTER}{self.student_id}", data=payload,
                                      timeout=timeout).json()
        if not register or "flag" not in register:
            raise RegistrationException("Bad request.")  # pragma: no cover
        if register["flag"] == "0":
            if "msg" in register:
                if register["msg"] == "所选教学班的上课时间与其他教学班有冲突！":
                    raise TimeConflictException
                else:  # pragma: no cover
                    raise RegistrationException(register["msg"])  # pragma: no cover
            else:
                raise RegistrationException("Unknown error.")  # pragma: no cover
        elif register["flag"] == "-1":
            raise FullCapacityException
        elif register["flag"] == "1":
            return
        else:
            raise RegistrationException(f"Unexpected response: {register}")  # pragma: no cover

    def _class_deregister(self, _class: SelectionClass, timeout=10):
        payload = {
            "kch_id": _class.internal_course_id,
            "jxb_ids": _class.register_id
        }
        deregister = self._session.post(f"{consts.SELECTION_DEREGISTER}{self.student_id}", data=payload,
                                        timeout=timeout).json()
        if deregister == "1":
            return
        elif deregister == "2":  # pragma: no cover
            raise DeregistrationException("Server busy.")  # pragma: no cover
        elif deregister == "3":  # pragma: no cover
            raise DeregistrationException("Unknown error.")  # pragma: no cover
        elif deregister == "4":  # pragma: no cover
            raise DeregistrationException("Illegal access.")  # pragma: no cover
        elif deregister == "5":  # pragma: no cover
            raise DeregistrationException("Validation failure.")  # pragma: no cover
        else:
            raise DeregistrationException(f"Unexpected response: {deregister}")  # pragma: no cover

    def _fetch_selection_classes(self, sector: SelectionSector, internal_course_id: str) -> List[dict]:
        payload = {
            **SelectionSectorSchema().dump(sector),
            **SelectionSharedInfoSchema().dump(sector.shared_info),
            "kch_id": internal_course_id
        }
        classes_query = self._session.post(f"{consts.SELECTION_QUERY_CLASSES}{self.student_id}", data=payload).json()
        return SelectionClassSchema(many=True).load(classes_query)

    def _fetch_selection_class(self, selection_class: SelectionClass):
        class_dicts = self._fetch_selection_classes(selection_class.sector, selection_class.internal_course_id)
        for class_dict in class_dicts:
            if class_dict["class_id"] == selection_class.class_id:
                return class_dict
        raise SelectionClassFetchException("Unable to fetch selection class information.")  # pragma: no cover

    def _get_selection_classes(self, sector: SelectionSector) -> List[SelectionClass]:
        payload = {
            **SelectionSectorSchema().dump(sector),
            **SelectionSharedInfoSchema().dump(sector.shared_info),
            "kspage": 1,
            "jspage": 5000
        }
        courses_query = self._session.post(f"{consts.SELECTION_QUERY_COURSES}{self.student_id}", data=payload).json()
        selection_classes: List[SelectionClass] = [SelectionClass(**item) for item in
                                                   SelectionCourseSchema(many=True).load(courses_query["tmpList"])]
        for _class in selection_classes:
            _class.sector = sector
            _class._load_func = partial(self._fetch_selection_class, _class)
            _class.is_registered = partial(self._class_is_registered, _class)
            _class.register = partial(self._class_register, _class)
            _class.deregister = partial(self._class_deregister, _class)
        return selection_classes

    @property
    def course_selection_sectors(self) -> List[SelectionSector]:
        """
        In iSJTU, courses are split into different sectors when selecting course.
        This property contains all available course sectors in this round of selection.
        """
        sectors_query = self._session.get(f"{consts.SELECTION_ALL_SECTORS_PARAM_URL}{self.student_id}").text
        if "对不起，当前不属于选课阶段" in sectors_query:
            raise SelectionNotAvailableException

        raw_shared_info = parse_shared_info(sectors_query)
        shared_info: SelectionSharedInfo = SelectionSharedInfoSchema().load(raw_shared_info)

        raw_sectors = parse_sectors(sectors_query)
        sectors = []
        for kklxdm, xkkz_id, name in raw_sectors:
            sector_query = self._session.post(f"{consts.SELECTION_SECTOR_PARAM_URL}{self.student_id}",
                                              data={"xkkz_id": xkkz_id,
                                                    "xszxzt": shared_info.self_selecting_status,
                                                    "kspage": 0, "jspage": 0}).text
            raw_sector = parse_sector(sector_query)
            sector: SelectionSector = SelectionSectorSchema().load(raw_sector)
            sector.name, sector.course_type_code, sector.xkkz_id, sector.shared_info = \
                name, kklxdm, xkkz_id, shared_info
            sector._func_classes = partial(self._get_selection_classes, sector=sector)
            sectors.append(sector)

        return sectors

    def flush_selection_class_cache(self):
        self._fetch_selection_classes.cache_clear()



import datetime
from dataclasses import dataclass
from typing import List, Optional

from pysjtu.models.base import Result, Results


@dataclass
class Exam(Result):
    """
    A model which describes an exam. Some fields may be empty.

    :param name: name of the course on which you are being examined.
    :type name: str
    :param location: the place where this exam is held.
    :type location: str
    :param seat: seat number
    :type seat: int
    :param course_id: course id of the course on which you are being examined.
    :type course_id: str
    :param course_name: course name of the course on which you are being examined.
    :type course_name: str
    :param class_name: class name of the class you are attending on the course which are being examined.
    :type class_name: str
    :param rebuild: whether this exam is a rebuild test.
    :type rebuild: bool
    :param credit: credits that the course provides.
    :type credit: float
    :param self_study: whether this course is a self study course.
    :type self_study: bool
    :param date: date of the exam
    :type date: datetime.date
    :param time: time range of the exam
    :type time: List[datetime.time]
    """
    name: str
    location: Optional[str] = None
    seat: Optional[int] = None
    course_id: Optional[str] = None
    course_name: Optional[str] = None
    class_name: Optional[str] = None
    rebuild: Optional[bool] = None
    credit: Optional[float] = None
    self_study: Optional[bool] = None
    date: Optional[datetime.date] = None
    time: Optional[List[datetime.time]] = None

    def __repr__(self):
        date_out = self.date.strftime("%Y-%m-%d")
        time_out = [time.strftime("%H:%M") for time in self.time]
        return f"<Exam \"{self.name}\" course_name={self.course_name} location={self.location} " \
               f"datetime={date_out}({time_out[0]}-{time_out[1]})>"


from pysjtu.schemas.exam import ExamSchema


class Exams(Results[Exam]):
    """
    A list-like interface to Exam collections.
    An additional filter method has been added to make filter operations easier.
    """
    _schema = ExamSchema
    _result_model = Exam



from dataclasses import dataclass
from enum import Enum
from typing import List

from pysjtu.models.base import Result


class LogicEnum(Enum):
    """ Used by :class:`GPAQueryParams` to specify condition logic. """
    AND = 0
    OR = 1


class CourseRange(Enum):
    """ Used by :class:`GPAQueryParams` to specify courses taken into account when ranking """
    ALL = "qbkc"
    CORE = "hxkc"


class Ranking(Enum):
    """ Used by :class:`GPAQueryParams` to specify student range upon which to rank """
    GRADE_AND_FIELD = "njzy"


@dataclass
class GPAQueryParams(Result):
    """
    A model which describes GPA query parameters. Used when performing gpa queries (pysjtu.Client().query_gpa(...)).
    You may leave fields empty if you don't want to filter by them.

    :param start_term: begin term of the query.
    :type start_term: int
    :param end_term: end term of the query.
    :type end_term: int
    :param condition_logic: logic applied between `has_roll`, `registered` and `attending`.
    :type condition_logic: :class:`LogicEnum`
    :param makeup_as_60: treat makeup scores (P) as 60.
    :type makeup_as_60: bool
    :param rebuild_as_60: treat rebuild scores (P) as 60.
    :type rebuild_as_60: bool
    :param gp_round: round gp to a given precision in decimal digits.
    :type gp_round: int
    :param gpa_round: round gpa to a given precision in decimal digits.
    :type gpa_round: int
    :param exclude_gp: exclude courses matching given criteria when calculating gp.
    :type exclude_gp: str
    :param exclude_gpa: exclude courses matching given criteria when calculating gpa.
    :type exclude_gpa: str
    :param course_whole: unknown parameter.
    :type course_whole: List[str]
    :param course_range: courses taken into account when ranking
    :type course_range: :class:`CourseRange`
    :param ranking: student range upon which to rank
    :type ranking: :class:`Ranking`
    :param has_roll: only include students who are enrolled in school.
    :type has_roll: bool
    :param registered: only include students who are registered.
    :type registered: bool
    :param attending: only include students who are attending school now.
    :type attending: bool
    """
    start_term: int
    end_term: int
    condition_logic: LogicEnum
    makeup_as_60: bool
    rebuild_as_60: bool
    gp_round: int
    gpa_round: int
    exclude_gp: str
    exclude_gpa: str
    course_whole: List[str]
    course_range: CourseRange
    ranking: Ranking
    has_roll: bool
    registered: bool
    attending: bool

    def __repr__(self):
        return f"<GPAQueryParams {self.__dict__}>"


@dataclass
class GPA(Result):
    """
    A model which describes GP & GPA and rankings.

    :param total_score: summed score of all matched courses.
    :type total_score: int
    :param course_count: number of all matched courses.
    :type course_count: int
    :param fail_count: number of failed courses.
    :type fail_count: int
    :param total_credit: summed credit of all matched courses.
    :type total_credit: float
    :param acquired_credit: summed credit of passed courses.
    :type acquired_credit: float
    :param failed_credit: summed credit of failed courses.
    :type failed_credit: float
    :param pass_rate: the pass rate of all matched courses.
    :type pass_rate: float
    :param gp: summed gp of all matched courses.
    :type gp: float
    :param gp_ranking: ranking of the gp.
    :type gp_ranking: int
    :param gpa: gpa of all matched courses.
    :type gpa: float
    :param gpa_ranking: ranking of the gpa.
    :type gpa_ranking: int
    :param total_students: number of students participates in the ranking.
    :type total_students: int
    """
    total_score: int
    course_count: int
    fail_count: int
    total_credit: float
    acquired_credit: float
    failed_credit: float
    pass_rate: float
    gp: float
    gp_ranking: int
    gpa: float
    gpa_ranking: int
    total_students: int

    def __repr__(self):
        return f"<GPA gp={self.gp} {self.gp_ranking}/{self.total_students} " \
               f"gpa={self.gpa} {self.gpa_ranking}/{self.total_students}>"



LOGIN_POST_URL = "https://jaccount.sjtu.edu.cn/jaccount/ulogin"
CAPTCHA_URL = "https://jaccount.sjtu.edu.cn/jaccount/captcha"

DEFAULT_BASE_URL = "https://i.sjtu.edu.cn"

HOME_URL = "/xtgl/index_initMenu.html"
SCHEDULE_URL = "/kbcx/xskbcx_cxXsKb.html?gnmkdm=N2151"
LOGIN_URL = "/jaccountlogin"
LOGOUT_URL = "/logout"
SCORE_URL = "/cjcx/cjcx_cxDgXscj.html?doType=query&gnmkdm=N305005"
SCORE_DETAIL_URL = "/cjcx/cjcx_cxXsXmcjList.html?gnmkdm=N305007&su="
EXAM_URL = "/kwgl/kscx_cxXsksxxIndex.html?doType=query&gnmkdm=N358105&su="
CALENDAR_URL = "/xtgl/index_cxshjdAreaFive.html?localeKey=zh_CN&gnmkdm=index&su="
COURSELIB_URL = "/design/funcData_cxFuncDataList.html?func_widget_guid=DA1B5BB30E1F4CB99D1F6F526537777B&gnmkdm" \
                "=N219904&su= "
GPA_PARAMS_URL = "/cjgl/common_cxGnzdxxList.html?gnmkdm=N309130"
GPA_CALC_URL = "/cjpmtj/gpapmtj_tjGpapmtj.html?gnmkdm=N309131&su="
GPA_QUERY_URL = "/cjpmtj/gpapmtj_cxGpaxjfcxIndex.html?doType=query&gnmkdm=N309131&su="
PROFILE_URL = "/xsxxxggl/xsgrxxwh_cxXsgrxx.html?gnmkdm=N100801&layout=default&su="
SELECTION_ALL_SECTORS_PARAM_URL = "/xsxk/zzxkyzb_cxZzxkYzbIndex.html?gnmkdm=N253512&layout=default&su="
SELECTION_SECTOR_PARAM_URL = "/xsxk/zzxkyzb_cxZzxkYzbDisplay.html?gnmkdm=N253512&su="
SELECTION_QUERY_COURSES = "/xsxk/zzxkyzb_cxZzxkYzbPartDisplay.html?gnmkdm=N253512&su="
SELECTION_QUERY_CLASSES = "/xsxk/zzxkyzb_cxJxbWithKchZzxkYzb.html?gnmkdm=N253512&su="
SELECTION_REGISTER = "/xsxk/zzxkyzb_xkBcZyZzxkYzb.html?gnmkdm=N253512&su="
SELECTION_IS_REGISTERED = "/xsxk/zzxkyzb_xkJcInXksjZzxkYzb.html?gnmkdm=N253512&su="
SELECTION_DEREGISTER = "/xsxk/zzxkyzb_tuikBcZzxkYzb.html?gnmkdm=N253512&su="

HEADERS = {"User-Agent": "Mozilla/5.0 (X11; Linux x86_64; rv:71.0) Gecko/20100101 Firefox/71.0",
           "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8",
           "Accept-Language": "zh-CN,zh;q=0.9,en-US;q=0.8,en;q=0.7,zh-TW;q=0.6"}
TERMS = [3, 12, 16]
CHINESE_WEEK = {"日": 0, "一": 1, "二": 2, "三": 3, "四": 4, "五": 5, "六": 6}



from io import BytesIO

from pkg_resources import resource_filename
import httpx

from pysjtu.utils import range_in_set
from pysjtu.exceptions import OCRException

try:
    import onnxruntime as rt  # type: ignore
    has_onnx = True
except ModuleNotFoundError:
    has_onnx = False

class Recognizer:
    """ Base class for Recognizers """

    def recognize(self, img: bytes):
        raise NotImplementedError  # pragma: no cover


class JCSSRecognizer(Recognizer):
    def __init__(self, url: str = "https://jcss.lightquantum.me"):
        self.url = url
        self.client = httpx.Client()

    def recognize(self, img: bytes):
        try:
            r = self.client.post(self.url, files={"image": BytesIO(img)})
            resp = r.json()
            return resp["data"]["prediction"]
        except Exception as e:  # pragma: no cover
            raise OCRException from e   # pragma: no cover


class LegacyRecognizer(Recognizer):
    """
    An legacy captcha recognizer.

    It first applies projection-based algorithm to the input image, then use a pre-trained shallow model
    to predict the answer.

    It's memory and cpu efficient. By using the provided model, the accuracy is around 90%.

    Usage::

        >>> import pysjtu
        >>> s = pysjtu.Session(ocr=LegacyRecognizer("svm_model.onnx"))
        >>> s.login('user@sjtu.edu.cn', 'something_secret')

    :param model_file: Pretrained ONNX model file.
    """

    def __init__(self, model_file: str = None):
        if not has_onnx:
            raise RuntimeError("Missing dependency: ONNXRuntime")
        if not model_file:
            model_file = resource_filename("pysjtu.ocr", "svm_model.onnx")
        self._clr = rt.InferenceSession(model_file)
        self._table = [0] * 156 + [1] * 100

    @staticmethod
    def row_not_empty(img, row):
        """
        A helper function to detect whether there's any black block in the specific row in a binary image.

        :param img: input image.
        :param row: specify a row to detect.

        :return a bool indicating the result.
        """
        for i in range(img.width):
            if img.getpixel((i, row)) != 1:
                return True
        return False

    @staticmethod
    def col_not_empty(img, col):
        """
        A helper function to detect whether there's any black block in the specific col in a binary image.

        :param img: input image.
        :param col: specify a col to detect.

        :return a bool indicating the result.
        """
        for i in range(img.height):
            if img.getpixel((col, i)) != 1:
                return True
        return False

    @staticmethod
    def h_split(img):
        """
        A helper function to split captcha into segments horizontally using projection.

        :param img: input image.

        :return segmented images.
        """
        col_filled = {col if LegacyRecognizer.col_not_empty(img, col) else None for col in range(img.width)}
        col_filled.remove(None)

        segments = range_in_set(col_filled)
        rtn = []
        for segment in segments:
            rtn.append(img.crop((segment.start, 0, segment.stop - 1, img.height)))
        return rtn

    @staticmethod
    def v_split(img):
        """
        A helper function to crop images vertically.

        :param img: input image.

        :return cropped images.
        """
        row_filled = {row if LegacyRecognizer.row_not_empty(img, row) else None for row in range(img.height)}
        row_filled.remove(None)
        segments = list(range_in_set(row_filled))
        top = min(segment.start for segment in segments)
        bottom = max(segment.stop for segment in segments)
        return img.crop((0, top, img.width, bottom))

    @staticmethod
    def normalize(img):
        """
        A helper function to normalize segmented & cropped image.

        :param img: input image.

        :return normalized image.
        """
        from PIL import Image

        rtn = Image.new("1", (20, 20), color=1)

        rtn.paste(img, (int((20 - img.width) / 2), int((20 - img.height) / 2)))
        return rtn

    def recognize(self, img: bytes):
        """
        Predict the captcha.

        :param img: A bytes array containing the captcha image.
        :return: captcha in plain text.
        """
        import numpy as np
        from PIL import Image

        img_rec = Image.open(BytesIO(img))
        img_rec = img_rec.convert("L")
        img_rec = img_rec.point(self._table, "1")

        segments = [self.normalize(self.v_split(segment)).convert("L").getdata() for segment in
                    self.h_split(img_rec)]

        np_segments = [np.array(segment, dtype=np.float32) for segment in segments]
        predicts = [self._clr.run(None, {self._clr.get_inputs()[0].name: np_segment}) for np_segment in np_segments]
        return "".join(str(predict[0][0]) for predict in predicts)


class NNRecognizer(Recognizer):
    """
    A Neural Network based captcha recognizer.

    It feeds the image directly into a pre-trained deep model to predict the answer.

    It consumes more memory and computing power than :class:`LegacyRecognizer`. By using the provided model,
    the accuracy is around 99%.

    It's currently the default recognizer.

    Usage::

        >>> import pysjtu
        >>> s = pysjtu.Session(ocr=NNRecognizer("nn_model.onnx"))
        >>> s.login('user@sjtu.edu.cn', 'something_secret')

    :param model_file: Pretrained ONNX model file.
    """

    def __init__(self, model_file: str = None):
        if not has_onnx:
            raise RuntimeError("Missing dependency: ONNXRuntime")
        if not model_file:
            model_file = resource_filename("pysjtu.ocr", "nn_model.onnx")
        self._table = [0] * 156 + [1] * 100
        self._sess = rt.InferenceSession(model_file)

    @staticmethod
    def _tensor_to_captcha(tensors):
        import numpy as np

        captcha = ""
        for tensor in tensors:
            asc = int(np.argmax(tensor, 1))
            if asc < 26:
                captcha += chr(ord("a") + asc)
        return captcha

    def recognize(self, img: bytes):
        """
        Predict the captcha.

        :param img: A bytes array containing the captcha image.
        :return: captcha in plain text.
        """
        import numpy as np
        from PIL import Image

        img_rec = Image.open(BytesIO(img))
        img_rec = img_rec.convert("L")
        img_rec = img_rec.point(self._table, "1")
        img_np = np.array(img_rec, dtype=np.float32)
        img_np = np.expand_dims(img_np, 0)
        img_np = np.expand_dims(img_np, 0)

        out_tensor = self._sess.run(None, {self._sess.get_inputs()[0].name: img_np})
        output = NNRecognizer._tensor_to_captcha(out_tensor)
        return output



import time
from typing import Union

from httpx.config import (TimeoutTypes, UNSET, UnsetType)

from pysjtu import consts
from pysjtu import models
from pysjtu.client.base import BaseClient


class ExamMixin(BaseClient):
    def __init__(self):
        super().__init__()

    def exam(self, year: int, term: int, timeout: Union[TimeoutTypes, UnsetType] = UNSET) \
            -> models.Results[models.Exam]:
        """
        Fetch your exams schedule of specific year & term.

        :param year: year for the new :class:`Exams` object.
        :param term: term for the new :class:`Exams` object.
        :param timeout: (optional) How long to wait for the server to send data before giving up.
        :return: A new :class:`Exams` object.
        :rtype: :class:`Exams`
        """
        raw = self._session.post(consts.EXAM_URL + str(self.student_id),
                                 data={"xnm": year, "xqm": consts.TERMS[term], "_search": False, "ksmcdmb_id": '',
                                       "kch": '', "kc": '', "ksrq": '', "kkbm_id": '',
                                       "nd": int(time.time() * 1000), "queryModel.showCount": 15,
                                       "queryModel.currentPage": 1, "queryModel.sortName": "",
                                       "queryModel.sortOrder": "asc", "time": 1}, timeout=timeout)
        scores = models.Exams(year, term)
        scores.load(raw.json()["items"])  # type: ignore
        return scores



from collections.abc import Generator
from os import path

import pytest

from pysjtu.utils import elfhash, flatten, has_callable, overlap, parse_slice, range_in_set, range_list_to_str, \
    replace_keys, schema_post_loader

DATA_DIR = path.join(path.dirname(path.abspath(__file__)), 'data')


def test_elfhash():
    assert elfhash("hello world") == 224648685
    assert elfhash("生活就像海洋，只有意志坚强的人才能到达彼岸。"
                   "This is an apple. I like apples. Apples are good for our health.") == 238480205


def test_parse_slice():
    class DummyObj1:
        __index__ = 0

    class DummyObj2:
        def __index__(self):
            return 0

    assert parse_slice(3) == 3
    assert parse_slice(DummyObj2()) == 0
    assert parse_slice(None) is None
    with pytest.raises(AttributeError):
        parse_slice(1.5)
    with pytest.raises(AttributeError):
        parse_slice(DummyObj1())


def test_has_callable():
    class Test:
        test_var1 = 0

        def __init__(self):
            self.test_var2 = 0

        def test_func(self):
            return 0

    not_a_class = 0
    assert has_callable(Test, "test_func")
    assert not has_callable(Test, "test_var1")
    assert not has_callable(Test, "test_var2")
    assert not has_callable(not_a_class, "wrong_func")


@pytest.mark.parametrize("base_dict", [{"key_1": "value_1", "key_2": "value_2", "key_3": "value_3"}])
@pytest.mark.parametrize("pairs, expected", [
    ([("key_1", "key_4")], {"key_4": "value_1", "key_2": "value_2", "key_3": "value_3"}),
    ([("key_1", "key_4"), ("key_2", "key_5")], {"key_4": "value_1", "key_5": "value_2", "key_3": "value_3"}),
    ([("key_2", "key_3"), ("key_3", "key_1")], {"key_1": "value_2"})
])
def test_replace_keys(base_dict, pairs, expected):
    test_dict = base_dict.copy()
    rtn_dict = replace_keys(test_dict, pairs)
    assert test_dict == rtn_dict == expected


def test_schema_post_loader():
    class DummySchema:
        def __init__(self, **kwargs):
            self.many = kwargs["many"] if "many" in kwargs.keys() else False

        def load(self, data):
            return self.many, data

    assert schema_post_loader(DummySchema, [1, 2]) == (True, [1, 2])
    assert schema_post_loader(DummySchema, {"k": "v"}) == (False, {"k": "v"})
    with pytest.raises(TypeError):
        schema_post_loader(DummySchema, 0)


def test_range_list_to_str():
    assert range_list_to_str([1]) == "1"
    assert range_list_to_str([1, range(3, 5)]) == "1或3或4"
    with pytest.raises(TypeError):
        range_list_to_str(1)


def test_range_in_set():
    assert len(list(range_in_set({}))) == 0

    ranges = range_in_set({1, 3, 4, 5, 7, 8, 9, 10})
    assert isinstance(ranges, Generator)
    assert list(ranges) == [range(1, 2), range(3, 6), range(7, 11)]


@pytest.mark.parametrize("objs, intersection", [
    ((2, range(1, 4)), {2}),
    (([1, 3, 4, 7, range(9, 11)], [range(2, 6), 10]), {3, 4, 10})
])
def test_overlap(objs, intersection):
    assert overlap(*objs) == intersection


@pytest.mark.parametrize("obj, flattened", [
    ([1, 3, range(5, 7)], [1, 3, 5, 6]),
    ([range(1, 7, 2)], [1, 3, 5])
])
def test_flatten(obj, flattened):
    assert isinstance(flatten(obj), Generator)
    assert list(flatten(obj)) == flattened



class LoadWarning(UserWarning):
    """ Missing fields in the loaded dict, which may cause a SessionException later. """


class DumpWarning(UserWarning):
    """ Missing fields in the dumped dict, which may cause a SessionException later. """


class GPACalculationException(Exception):
    """ A failure has been reported by the remote server when calculating GPA. """


class SelectionNotAvailableException(Exception):
    """ Course selection function is not available at this time. """


class SelectionClassFetchException(Exception):
    """ Unable to fetch selection class information. """


class RegistrationException(Exception):
    """ Base exception for class registration failures. """


class DeregistrationException(Exception):
    """ Base exception for class deregistration failures. """


class FullCapacityException(RegistrationException):
    """ There's no room for this class. """


class TimeConflictException(RegistrationException):
    """ There's a time conflict when registering for this class. """


class SessionException(Exception):
    """ The session is expired or invalid, and we can't renew it automatically. """


class LoginException(Exception):
    """ An error occurred when logging into the website """


class OCRException(Exception):
    """ An error occurred when solving captcha """


class ServiceUnavailable(Exception):
    """ The website is down or under maintenance. """



import typing

from marshmallow import EXCLUDE, Schema, fields, post_load  # type: ignore

from pysjtu.schemas.base import CourseTime, CourseWeek, SplitField


class CreditHourDetail(fields.Field):
    def _deserialize(
            self,
            value: typing.Any,
            attr: typing.Optional[str],
            data: typing.Optional[typing.Mapping[str, typing.Any]],
            **kwargs
    ):
        if not value:
            return None  # pragma: no cover
        class_hour_details = value.split(",")
        rtn = {}
        for item in class_hour_details:
            try:
                name, hour = item.split(":")
                rtn[name] = float(hour)
            except ValueError:
                rtn["N/A"] = 0
        return rtn


class ScheduleCourseSchema(Schema):
    class Meta:
        unknown = EXCLUDE

    name = fields.Str(required=True, data_key="kcmc")
    day = fields.Int(data_key="xqj")
    week = CourseWeek(data_key="zcd")
    time = CourseTime(data_key="jcs")
    location = fields.Str(data_key="cdmc")
    credit = fields.Float(data_key="xf")
    assessment = fields.Str(data_key="khfsmc")
    remark = fields.Str(data_key="xkbz")
    teacher_name = SplitField(data_key="xm", sep=",")
    teacher_title = SplitField(data_key="zcmc", sep=",")
    course_id = fields.Str(required=True, data_key="kch_id")
    class_name = fields.Str(required=True, data_key="jxbmc")
    class_id = fields.Str(required=True, data_key="jxb_id")
    hour_total = fields.Int(data_key="zxs")
    hour_remark = CreditHourDetail(data_key="kcxszc")
    hour_week = fields.Int(data_key="zhxs")
    field = fields.Str(data_key="zyfxmc")

    # noinspection PyUnusedLocal
    @post_load
    def wrap(self, data, **kwargs):
        return ScheduleCourse(**data)


from pysjtu.models.schedule import ScheduleCourse



from pysjtu import consts
from pysjtu.client.base import BaseClient
from pysjtu.models.profile import Profile
from pysjtu.parser.profile import parse
from pysjtu.schemas.profile import profile_fields


class ProfileMixin(BaseClient):
    def __init__(self):
        super().__init__()

    @property
    def profile(self) -> Profile:
        """ Get the user profile of the current session. """
        if "profile" not in self._session._cache_store:
            rtn = self._session.get(f"{consts.PROFILE_URL}{self.student_id}")
            self._session._cache_store["profile"] = parse(profile_fields, rtn.text)
        return Profile(**self._session._cache_store["profile"])



import pickle
from datetime import date
from functools import partial
from tempfile import NamedTemporaryFile

import httpx
import pytest
import respx

from pysjtu.client import Client, create_client
from pysjtu.exceptions import DumpWarning, GPACalculationException, LoadWarning, LoginException, ServiceUnavailable, \
    SessionException, SelectionNotAvailableException, TimeConflictException, FullCapacityException
from pysjtu.models import CourseRange, Exams, GPA, GPAQueryParams, LogicEnum, QueryResult, Schedule, Scores, Profile
from pysjtu.ocr import JCSSRecognizer
from pysjtu.session import BaseSession, Session
from .mock_server import app
from dataclasses import asdict


@pytest.fixture
def logged_session(mocker):
    mocker.patch.object(JCSSRecognizer, "recognize", return_value="ipsum")
    sess = Session(_mocker_app=app, retry=[0], timeout=1)
    sess.login("FeiLin", "WHISPERS")
    return sess


@pytest.fixture
def check_login():
    def _check_login(session):
        return "519027910001" in session.get("https://i.sjtu.edu.cn/xtgl/index_initMenu.html").text

    return _check_login


class TestSession:
    @respx.mock
    def test_base_url(self):
        respx.get("https://i.sjtu.edu.cn/test", content="1")
        respx.get("https://kbcx.sjtu.edu.cn/test", content="2")
        sess = Session()
        assert sess.base_url == "https://i.sjtu.edu.cn"
        assert sess.get("/test").text == "1"
        sess_2 = Session(base_url="https://kbcx.sjtu.edu.cn")
        assert sess_2.base_url == "https://kbcx.sjtu.edu.cn"
        assert sess_2.get("/test").text == "2"

    @respx.mock
    def test_secure_req(self):
        respx.get("http://secure.page.edu.cn", content=httpx.ConnectionClosed())
        respx.get("http://secure.page.edu.cn:8889", content=httpx.ConnectionClosed())
        respx.get("https://fail.page.edu.cn", content=httpx.ConnectionClosed())
        respx.get("https://secure.page.edu.cn")
        sess = Session()

        resp = sess._secure_req(partial(httpx.get, "http://secure.page.edu.cn"))
        assert resp.status_code == 200

        resp = sess._secure_req(partial(httpx.get, "http://secure.page.edu.cn:8889"))
        assert resp.status_code == 200

        with pytest.raises(httpx.exceptions.NetworkError):
            sess._secure_req(partial(httpx.get, "https://fail.page.edu.cn"))

    def test_context(self, mocker):
        tmpfile = NamedTemporaryFile()
        pickle.dump({"username": "FeiLin", "password": "WHISPERS"}, tmpfile)
        tmpfile.seek(0)

        mocker.patch.object(JCSSRecognizer, "recognize", return_value="ipsum")
        with Session(_mocker_app=app, session_file=tmpfile.file):
            pass
        tmpfile.seek(0)

        assert pickle.load(tmpfile)["cookies"]

    def test_init(self, mocker, check_login):
        tmpfile = NamedTemporaryFile()
        mocker.patch.object(JCSSRecognizer, "recognize", return_value="ipsum")
        sess = Session(_mocker_app=app, username="FeiLin", password="WHISPERS")
        assert check_login(sess)
        cookie = sess.cookies
        sess.dump(tmpfile.file)
        tmpfile.seek(0)

        with pytest.warns(LoadWarning):
            sess = Session(_mocker_app=app, cookies=cookie)
            assert check_login(sess)

        sess = Session(_mocker_app=app, session_file=tmpfile.file)
        assert check_login(sess)

    def test_req(self, logged_session, check_login):
        with pytest.raises(ServiceUnavailable):
            logged_session.get("https://i.sjtu.edu.cn/503")

        logged_session.get("https://i.sjtu.edu.cn/expire_me")
        assert logged_session.get("https://i.sjtu.edu.cn/xtgl/index_initMenu.html",
                                  validate_session=False).url.full_path == "/xtgl/login_slogin.html"
        with pytest.raises(SessionException):
            logged_session.get("https://i.sjtu.edu.cn/xtgl/index_initMenu.html", auto_renew=False)
        assert check_login(logged_session)

        logged_session.get("https://i.sjtu.edu.cn/expire_me")
        logged_session._username = None
        with pytest.raises(SessionException):
            logged_session.get("https://i.sjtu.edu.cn/xtgl/index_initMenu.html")

        with pytest.raises(httpx.exceptions.HTTPError):
            logged_session.get("https://i.sjtu.edu.cn/404")

    def test_req_methods(self, logged_session):
        assert logged_session.get("https://i.sjtu.edu.cn/ping").text == "pong"
        logged_session.head("https://i.sjtu.edu.cn/ping")
        assert logged_session.post("https://i.sjtu.edu.cn/ping", data="lorem ipsum").text == "lorem ipsum"
        assert logged_session.patch("https://i.sjtu.edu.cn/ping").text == "pong"
        assert logged_session.put("https://i.sjtu.edu.cn/ping").text == "pong"
        assert logged_session.delete("https://i.sjtu.edu.cn/ping").text == "pong"
        options = logged_session.options("https://i.sjtu.edu.cn/ping").headers["allow"]
        assert sorted(options.split(", ")) == ['DELETE', 'GET', 'HEAD', 'OPTIONS', 'PATCH', 'POST', 'PUT']

    @respx.mock
    def test_req_partial_url(self):
        respx.get("http://dummy.url/test_path", content="pass")
        assert Session(base_url="http://dummy.url").get("/test_path").text == "pass"

    def test_login(self, logged_session, check_login):
        assert check_login(logged_session)

        with pytest.raises(LoginException):
            logged_session.login("Cookie☆", "1145141919810")
            logged_session.login("Cookie☆", "buriburiburi")

    def test_logout(self, logged_session, check_login):
        logged_session.logout(purge_session=False)
        assert logged_session.get("https://i.sjtu.edu.cn/is_login").text == "False"
        assert check_login(logged_session)

        logged_session.logout()
        with pytest.raises(SessionException):
            logged_session.get("https://i.sjtu.edu.cn/xtgl/index_initMenu.html")

    def test_loads_dumps(self, logged_session, check_login):
        cookie = logged_session.cookies
        dumps = logged_session.dumps()

        sess = Session(_mocker_app=app)
        sess.loads({"username": "FeiLin", "password": "WHISPERS"})
        assert check_login(sess)

        with pytest.warns(LoadWarning):
            sess.loads({})
        assert sess.cookies == httpx.Cookies({})
        assert not sess._username
        assert not sess._password

        sess = Session(_mocker_app=app)
        with pytest.raises(TypeError):
            sess.loads({"cookies": "Cookie☆"})
        with pytest.warns(LoadWarning):
            sess.loads({"cookies": {}})
            sess.loads({"cookies": cookie})
        assert check_login(sess)
        with pytest.warns(DumpWarning):
            sess.dumps()

        sess = Session(_mocker_app=app)
        sess.loads(dumps)
        assert check_login(sess)

        # test auto renew mechanism
        logged_session.logout()
        sess = Session(_mocker_app=app)
        sess.loads(dumps)
        assert check_login(sess)

    def test_load_dump(self, logged_session, check_login, tmp_path):
        tmp_file = NamedTemporaryFile()
        logged_session.dump(tmp_file.file)
        tmp_file.seek(0)
        sess = Session(_mocker_app=app)
        sess.load(tmp_file.file)
        assert check_login(sess)

        tmp_file = tmp_path / "tmpfile_1"
        # noinspection PyTypeChecker
        open(tmp_file, mode="a").close()
        logged_session.dump(tmp_file)
        sess = Session(_mocker_app=app)
        sess.load(tmp_file)
        assert check_login(sess)

        tmp_file = str(tmp_path / "tmpfile_2")
        open(tmp_file, mode="a").close()
        logged_session.dump(tmp_file)
        sess = Session(_mocker_app=app)
        sess.load(tmp_file)
        assert check_login(sess)

        with pytest.raises(TypeError):
            # noinspection PyTypeChecker
            sess.load(0)
        with pytest.raises(TypeError):
            # noinspection PyTypeChecker
            sess.dump(0)

        empty_file = NamedTemporaryFile()
        sess = Session(_mocker_app=app)
        with pytest.warns(LoadWarning):
            sess.load(empty_file.file)

        empty_file = tmp_path / "empty_file"
        # noinspection PyTypeChecker
        open(empty_file, mode="a").close()
        sess = Session(_mocker_app=app)
        with pytest.warns(LoadWarning):
            sess.load(empty_file)

    def test_properties(self, logged_session):
        cookie = logged_session.cookies

        sess = Session(_mocker_app=app)
        assert sess.proxies == {}

        assert isinstance(sess.timeout, httpx.Timeout)
        sess.timeout = httpx.Timeout(1.0)
        sess.timeout = 1
        sess.timeout = (1, 5)
        with pytest.raises(TypeError):
            sess.timeout = "1"

        assert isinstance(sess.cookies, httpx.Cookies)
        with pytest.raises(SessionException):
            sess.cookies = {}
        sess._cookies = {}
        sess._cache_store = {"key": "value"}
        sess.cookies = cookie
        assert sess._cache_store == {}
        assert sess._cookies == sess.cookies


@pytest.fixture
def logged_client(logged_session):
    return Client(logged_session)


class TestClient:
    class DummySession(BaseSession):
        _cache_store = {}

        def get(self): ...

        def post(self): ...

    class DummySession2:
        def get(self): ...

        def post(self): ...

    # noinspection PyTypeChecker
    def test_init(self, logged_session):
        Client(logged_session)
        Client(self.DummySession())
        with pytest.raises(TypeError):
            Client(0)
        with pytest.raises(TypeError):
            Client(self.DummySession2())
        client = create_client("FeiLin", "WHISPERS", _mocker_app=app)
        assert client.student_id == 519027910001

    def test_student_id(self, logged_client):
        assert logged_client.student_id == 519027910001

    def test_term_start_date(self, logged_client):
        assert logged_client.term_start_date == date(2019, 9, 9)

    def test_gpa_query_params(self, logged_client):
        assert isinstance(logged_client.default_gpa_query_params, GPAQueryParams)

    def test_schedule(self, logged_client):
        schedule = logged_client.schedule(2019, 0)
        assert isinstance(schedule, Schedule)
        assert len(schedule) == 3

    def test_get_score(self, logged_client):
        score = logged_client.score(2019, 0)
        assert isinstance(score, Scores)
        assert len(score) == 3
        assert len(score[0].detail) == 2

    def test_exam(self, logged_client):
        exam = logged_client.exam(2019, 0)
        assert isinstance(exam, Exams)
        assert len(exam) == 3

    def test_course(self, logged_client):
        courses = logged_client.query_courses(2019, 0, name="高等数学", page_size=40)
        assert isinstance(courses, QueryResult)
        assert len(courses) == 90
        assert len(list(courses)) == 90

    def test_gpa_fail(self, logged_client):
        params = logged_client.default_gpa_query_params
        params.condition_logic = LogicEnum.AND
        with pytest.raises(GPACalculationException) as e:
            logged_client.gpa(params)
        assert str(e.value) == "Unauthorized."
        params.condition_logic = LogicEnum.OR
        params.course_range = CourseRange.ALL
        with pytest.raises(GPACalculationException) as e:
            logged_client.gpa(params)
        assert str(e.value) == "Calculation failure."
        params.course_range = CourseRange.CORE
        gpa = logged_client.gpa(params)
        assert isinstance(gpa, GPA)

    def test_selection(self, logged_client):
        with pytest.raises(SelectionNotAvailableException):
            sectors = logged_client.course_selection_sectors
        logged_client._session.get("/test_selection")
        for _ in range(3):
            sectors = logged_client.course_selection_sectors
        assert logged_client._session.get("get_session?key=query_all_sectors").text == "3"
        assert len(sectors) == 6

        for _ in range(3):
            classes = sectors[0].classes
        assert logged_client._session.get("get_session?key=query_courses").text == "1"
        _ = logged_client.course_selection_sectors[0].classes
        assert logged_client._session.get("get_session?key=query_courses").text == "2"

        # noinspection PyUnboundLocalVariable
        _class = classes[0]
        with pytest.raises(TimeConflictException):
            _class.register()
        logged_client._session.get("test_no_conflict")
        with pytest.raises(FullCapacityException):
            _class.register()
        logged_client._session.get("test_no_full")
        _class.register()
        assert _class.is_registered() is True
        _class.deregister()
        assert _class.is_registered() is False

        _ = _class.register_id
        assert logged_client._session.get("get_session?key=query_classes").text == "1"
        logged_client.flush_selection_class_cache()
        _ = logged_client.course_selection_sectors[0].classes[0].register_id
        assert logged_client._session.get("get_session?key=query_classes").text == "2"

    def test_profile(self, logged_client):
        profile_1 = logged_client.profile
        assert isinstance(profile_1, Profile)
        assert logged_client._session.get("get_session?key=query_profile").text == "1"

        profile_2 = logged_client.profile
        assert profile_1 == profile_2
        assert logged_client._session.get("get_session?key=query_profile").text == "1"



from dataclasses import dataclass
from typing import List, Optional

from pysjtu.models.base import Result


@dataclass
class LibCourse(Result):
    """
    A model which describes a course in CourseLib. Some fields may be empty.

    :param name: literal name of the course.
    :type name: str
    :param day: in which day(s) of weeks classes are given.
    :type day: int
    :param week: in which week(s) classes are given.
    :type week: list
    :param time: at which time of days classes are given.
    :type time: range
    :param location: the place where classes are given.
    :type location: str
    :param locations: the places where classes are given.
    :type locations: List[str]
    :param faculty: the faculty which offers this course.
    :type faculty: str
    :param credit: credits that the course provides.
    :type credit: float
    :param teacher: the teacher who offers this course.
    :type teacher: List[str]
    :param course_id: course id.
    :type course_id: str
    :param class_name: class name (constant between years).
    :type class_name: str
    :param class_id: class id (variable between years).
    :type class_id: str
    :param class_composition: students from which faculties do the course consists of.
    :type class_composition: List[str]
    :param hour_total: credit hours of the course.
    :type hour_total: int
    :param hour_remark: detailed explanation of the credit hours.
    :type hour_remark: dict
    :param seats: number of seats available in this course.
    :type seats: int
    :param students_elected: number of students elected this course.
    :type students_elected: int
    :param students_planned: number of students planned when setting this course.
    :type students_planned: int
    """
    name: str
    day: Optional[int] = None
    week: Optional[list] = None
    time: Optional[range] = None
    location: Optional[str] = None
    locations: Optional[List[str]] = None
    faculty: Optional[str] = None
    credit: Optional[float] = None
    teacher: Optional[List[str]] = None
    course_id: Optional[str] = None
    class_name: Optional[str] = None
    class_id: Optional[str] = None
    class_composition: Optional[List[str]] = None
    hour_total: Optional[int] = None
    hour_remark: Optional[dict] = None
    seats: Optional[int] = None
    students_elected: Optional[int] = None
    students_planned: Optional[int] = None

    def __repr__(self):
        return f"<LibCourse {self.name} class_name={self.class_name}>"



from .course import CourseLibMixin
from .exam import ExamMixin
from .gpa import GPAMixin
from .profile import ProfileMixin
from .schedule import ScheduleMixin
from .score import ScoreMixin
from .selection import SelectionMixin



import re
from functools import partial
from typing import Dict, List, Tuple

field_pattern = re.compile('id="(?P<k>.*?)" value="(?P<v>.*?)"/>')
sectors_pattern = re.compile("queryCourse\\(this,'(?P<kklxdm>\\d*)','(?P<xkkz_id>.*?)'.*>(?P<name>.*)</a>")


def parse_fields(html: str, fields: list) -> Dict[str, str]:
    """
    A helper function to extract args from hidden fields in html.

    :param html: Input html src.
    :param fields: Fields in need.
    :return: A dict contains specific args.

    :raises StopIteration: if fields are missing in the given html src.
    """
    result = {}
    _fields = fields.copy()

    param_iter = field_pattern.finditer(html)
    while _fields:
        k, v = next(param_iter).groups()
        if k in _fields:
            result[k] = v
            _fields.remove(k)
    return result


def parse_sectors(html: str) -> List[Tuple[str]]:
    """
    Extract available sectors from html.

    :param html: Input html src.
    :return: A list of tuples contains query parameters of all available sectors
    """
    return sectors_pattern.findall(html)


parse_shared_info = partial(parse_fields,
                            fields=["xqh_id", "zyh_id", "njdm_id", "bh_id", "xkxnm", "xkxqm", "xszxzt", "ccdm",
                                    "xslbdm", "xbm", "zyfx_id", "xsbj"])
parse_sector = partial(parse_fields, fields=["rwlx", "xkly", "tykczgxdcs", "bklx_id", "txbsfrl", "kkbk", "sfkknj",
                                             "sfkkzy", "sfznkx", "zdkxms"])



import typing
from datetime import datetime

from marshmallow import EXCLUDE, Schema, fields, post_load  # type: ignore

from pysjtu.schemas.base import ChineseBool


class ExamDate(fields.Field):
    def _deserialize(
            self,
            value: typing.Any,
            attr: typing.Optional[str],
            data: typing.Optional[typing.Mapping[str, typing.Any]],
            **kwargs
    ):
        if not value:
            return None  # pragma: no cover
        return datetime.strptime(value[:value.index("(")], "%Y-%m-%d").date()


class ExamTime(fields.Field):
    def _deserialize(
            self,
            value: typing.Any,
            attr: typing.Optional[str],
            data: typing.Optional[typing.Mapping[str, typing.Any]],
            **kwargs
    ):
        if not value:
            return None  # pragma: no cover
        raw_time = value[value.find("(") + 1:-1].split("-")
        return [datetime.strptime(time, "%H:%M").time() for time in raw_time]


class ExamSchema(Schema):
    class Meta:
        unknown = EXCLUDE

    name = fields.Str(required=True, data_key="ksmc")
    location = fields.Str(data_key="cdmc")
    seat = fields.Int(data_key="zwh")
    course_id = fields.Str(data_key="kch")
    course_name = fields.Str(data_key="kcmc")
    class_name = fields.Str(data_key="jxbmc")
    rebuild = ChineseBool(data_key="cxbj")
    credit = fields.Float(data_key="xf")
    self_study = ChineseBool(data_key="zxbj")
    date = ExamDate(data_key="kssj", load_only=True)
    time = ExamTime(data_key="kssj", load_only=True)

    # noinspection PyUnusedLocal
    @post_load
    def wrap(self, data, **kwargs):
        return Exam(**data)


from pysjtu.models.exam import Exam



import sys
from os import path

import pytest
import respx

from pysjtu.ocr import LegacyRecognizer, NNRecognizer, JCSSRecognizer

CAPTCHA_DIR = path.join(path.dirname(path.abspath(__file__)), 'resources/captcha')
MODEL_FILE = path.join(path.abspath("."), 'svm_model.onnx')
NN_MODEL_FILE = path.join(path.abspath("."), 'nn_model.onnx')


@pytest.mark.xfail(reason="SVM-based and ResNet-based recognizer doesn't recognize captcha correctly sometime.")
@pytest.mark.skipif(sys.version_info > (3, 8), reason="Python 3.9 doesn't have ONNXRuntime, yet.")
@pytest.mark.parametrize("captcha", [
    "mwgi", "qkxvx", "qrdxb", "rysv", "zwtco"
])
def test_recognizer(captcha):
    predictors = [LegacyRecognizer(), NNRecognizer()]
    for predictor in predictors:
        assert predictor.recognize(open(path.join(CAPTCHA_DIR, captcha + ".jpg"), mode="rb").read()) == captcha

@respx.mock
def test_jcss_recognizer():
    respx.post("https://jcss.lightquantum.me", content='{"status":"success","data":{"prediction":"gbmke","elapsed_time":2}}')
    predictor = JCSSRecognizer()
    assert predictor.recognize(b'fbkfbkfbk') == "gbmke"


from dataclasses import dataclass
from typing import List, Optional

from pysjtu.models.base import Result, Results


@dataclass
class ScheduleCourse(Result):
    """
    A model which describes a course in CourseLib. Some fields may be empty.

    :param name: literal name of the course.
    :type name: str
    :param day: in which day(s) of weeks classes are given.
    :type day: int
    :param week: in which week(s) classes are given.
    :type week: list
    :param time: at which time of days classes are given.
    :type time: range
    :param location: the place where classes are given.
    :type location: str
    :param credit: credits that the course provides.
    :type credit: int
    :param assessment: assessment method of this course. (exams, assesses, etc)
    :type assessment: str
    :param remark: remarks of this course.
    :type remark: str
    :param teacher_name: the teacher who offers this course.
    :type teacher_name: List[str]
    :param teacher_title: title of the course's teacher.
    :type teacher_title: List[str]
    :param course_id: course id.
    :type course_id: str
    :param class_name: class name (constant between years).
    :type class_name: str
    :param class_id: class id (variable between years).
    :type class_id: str
    :param hour_total: credit hours of the course.
    :type hour_total: int
    :param hour_remark: detailed explanation of the credit hours.
    :type hour_remark: dict
    :param hour_week: credit hours of the course every week.
    :type hour_week: int
    :param field: professional field of this course.
    :type field: str
    """
    name: str
    course_id: str
    class_name: str
    class_id: str
    day: Optional[int] = None
    week: Optional[list] = None
    time: Optional[range] = None
    location: Optional[str] = None
    credit: Optional[int] = None
    assessment: Optional[str] = None
    remark: Optional[str] = None
    teacher_name: Optional[List[str]] = None
    teacher_title: Optional[List[str]] = None
    hour_total: Optional[int] = None
    hour_remark: Optional[dict] = None
    hour_week: Optional[int] = None
    field: Optional[str] = None

    def __repr__(self):
        return f"<ScheduleCourse {self.name} week={self.week} day={self.day} time={self.time}>"


from pysjtu.schemas.schedule import ScheduleCourseSchema


class Schedule(Results[ScheduleCourse]):
    """
    A list-like interface to Schedule collections.
    An additional filter method has been added to make filter operations easier.
    """
    _schema = ScheduleCourseSchema
    _result_model = ScheduleCourse



from dataclasses import dataclass
from datetime import datetime
from typing import Optional

from .base import Gender


@dataclass(frozen=True)
class Profile:
    student_id: int
    name: str
    name_pinyin: Optional[str]
    former_name: Optional[str]
    gender: Gender
    certificate_type: str
    certificate_number: int
    birth_date: datetime
    enrollment_date: datetime
    birthplace: Optional[str]
    ethnicity: Optional[str]
    native_place: Optional[str]
    foreign_status: Optional[str]
    political_status: Optional[str]
    enrollment_province: Optional[str]
    nationality: Optional[str]
    domicile_place: Optional[str]
    cee_candidate_number: Optional[int]
    middle_school: Optional[str]
    religion: Optional[str]
    email: Optional[str]
    cellphone: Optional[int]
    family_address: Optional[str]
    mailing_address: Optional[str]
    landline: Optional[int]
    zip_code: Optional[int]



from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional

from lxml.html import HtmlElement, fromstring


@dataclass(frozen=True)
class ProfileField:
    name: str
    xpath: str
    post_parse: Optional[Callable[[str], Any]] = lambda x: x

    def parse(self, el: HtmlElement) -> Any:
        raw_fields = el.xpath(self.xpath)
        if not raw_fields:
            return None
        raw_field = raw_fields[0].text

        raw_text = raw_field.strip() if isinstance(raw_field, str) else None
        if not raw_text:
            return None

        # noinspection PyArgumentList
        return self.post_parse(raw_text.strip())


def parse(fields: List[ProfileField], src: str) -> Dict[str, Any]:
    el = fromstring(src)
    return {field.name: field.parse(el) for field in fields}



from datetime import date, time
from functools import partial
from math import ceil

import pytest

from pysjtu.models import QueryResult, GPAQueryParams, GPA, LibCourse, Exam, ScoreFactor, Score, ScheduleCourse, \
    Exams, Scores, Schedule, LazyResult, PARTIAL, SelectionClass, SelectionSector
from pysjtu.schemas import ExamSchema, ScoreSchema, ScheduleCourseSchema


@pytest.fixture
def dummy_req():
    class DummyResp:
        def __init__(self, resp):
            self._resp = resp

        def json(self):
            return self._resp

    class DummyReq:
        def __init__(self):
            self._is_called = False

        def __call__(self, total, data):
            assert data["validate"] == "Lorem Ipsum"
            count = data["queryModel.showCount"]
            page = data["queryModel.currentPage"]
            assert page <= ceil(total / count)

            self._is_called = True
            resp = {"totalResult": total, "items": range(count * (page - 1) + 1, count * page + 1)}
            return DummyResp(resp)

        @property
        def is_called(self):
            _tmp = self._is_called
            self._is_called = False
            return _tmp

    return DummyReq()


# noinspection PyStatementEffect
def test_query_result(dummy_req):
    def dummy_post_func(x):
        return {"post": x}

    result = QueryResult(partial(dummy_req, 200), dummy_post_func, {"validate": "Lorem Ipsum"}, page_size=10)

    assert len(result) == 200

    assert result[50] == {"post": 51}
    assert result[-2] == {"post": 199}
    assert result[-200] == {"post": 1}
    with pytest.raises(IndexError):
        result[200]
    with pytest.raises(IndexError):
        result[-201]
    with pytest.raises(TypeError):
        # noinspection PyTypeChecker
        result[1.5]

    assert result[:10] == {"post": list(range(1, 11))}
    assert result[:15] == {"post": list(range(1, 16))}
    assert result[185:] == {"post": list(range(186, 201))}
    assert result[:] == {"post": list(range(1, 201))}
    assert result[180:-100] == {"post": []}
    assert result[185:999] == result[185:]
    assert result[204:209] == {"post": []}
    assert result[-1:999] == {"post": [200]}
    assert list(result) == [{"post": item} for item in range(1, 201)]
    with pytest.raises(AttributeError):
        result[1.5:]
    with pytest.raises(AttributeError):
        result[:1.5]

    dummy_obj = dummy_req
    result = QueryResult(partial(dummy_obj, 200), dummy_post_func, {"validate": "Lorem Ipsum"}, page_size=10)
    result[5:25]
    assert dummy_obj.is_called
    result[4:24]
    assert not dummy_obj.is_called
    result.flush_cache()
    result[4:24]
    assert dummy_obj.is_called


def test_lazy_model(mocker):
    class DummyModel(LazyResult):
        normal_field: int = 0
        lazy_field_1: int = PARTIAL
        lazy_field_2: str = PARTIAL

    fake_load_func = mocker.Mock(return_value={"lazy_field_1": 1, "lazy_field_2": "2"})
    model = DummyModel()
    model._load_func = fake_load_func

    assert model.normal_field == 0
    fake_load_func.assert_not_called()
    fake_load_func.reset_mock()

    assert model.lazy_field_1 == 1
    fake_load_func.assert_called_once()
    fake_load_func.reset_mock()

    assert model.lazy_field_2 == "2"
    fake_load_func.assert_not_called()


@pytest.mark.parametrize("model, members, repr_pair", [
    (SelectionClass,
     ["name", "credit", "course_id", "internal_course_id", "class_name", "class_id", "students_registered",
      "register_id", "teachers", "locations", "time", "course_type", "remark", "students_planned", "sector"],
     ({"class_name": "Calculus", "name": "AA000-0"},
      "<SelectionClass Calculus AA000-0>")),

    (SelectionSector,
     ["task_type", "xkly", "pe_op_param", "sector_type_id", "txbsfrl", "kkbk", "course_type_code", "name", "xkkz_id",
      "include_other_grades", "include_other_majors", "sfznkx", "zdkxms", "shared_info"],
     ({"name": "major"},
      "<SelectionSector major>")),

    (GPAQueryParams,
     ["start_term", "end_term", "condition_logic", "makeup_as_60", "rebuild_as_60", "gp_round", "gpa_round",
      "exclude_gp", "exclude_gpa", "course_whole", "course_range", "ranking", "has_roll", "registered", "attending"],
     ({"start_term": 0},
      "<GPAQueryParams {'start_term': 0, 'end_term': None, 'condition_logic': None, 'makeup_as_60': None, "
      "'rebuild_as_60': None, 'gp_round': None, 'gpa_round': None, 'exclude_gp': None, 'exclude_gpa': None, "
      "'course_whole': None, 'course_range': None, 'ranking': None, 'has_roll': None, 'registered': None, "
      "'attending': None}>")),

    (GPA,
     ["total_score", "course_count", "fail_count", "total_credit", "acquired_credit", "failed_credit", "pass_rate",
      "gp",
      "gp_ranking", "gpa", "gpa_ranking", "total_students"],
     ({"gp": 80, "gp_ranking": 50, "gpa": 3.8, "gpa_ranking": 40, "total_students": 200},
      "<GPA gp=80 50/200 gpa=3.8 40/200>")),

    (LibCourse,
     ["name", "day", "week", "time", "location", "locations", "faculty", "credit", "teacher", "course_id", "class_name",
      "class_id", "class_composition", "hour_total", "hour_remark", "seats", "students_elected", "students_planned"],
     ({"name": "Calculus", "class_name": "AA001"}, "<LibCourse Calculus class_name=AA001>")),

    (Exam,
     ["name", "location", "course_id", "course_name", "class_name", "rebuild", "credit", "self_study", "date", "time"],
     ({"name": "Calculus Final", "course_name": "Calculus",
       "location": "LocDummy", "date": date(2012, 12, 21), "time": [time(13, 0), time(15, 0)]},
      "<Exam \"Calculus Final\" course_name=Calculus location=LocDummy datetime=2012-12-21(13:00-15:00)>")),

    (ScoreFactor,
     ["name", "percentage", "score"],
     ({"name": "Exam", "percentage": 0.6, "score": 92}, "<ScoreFactor Exam(60.0%)=92>")),

    (Score,
     ["name", "teacher", "score", "credit", "gp", "invalid", "course_type", "category", "score_type", "method",
      "course_id", "class_name", "class_id"],
     ({"name": "Calculus", "score": 87, "credit": 6.0, "gp": 3.7}, "<Score Calculus score=87 credit=6.0 gp=3.7>")),

    (ScheduleCourse,
     ["name", "day", "week", "time", "location", "credit", "assessment", "remark", "teacher_name", "teacher_title",
      "course_id", "class_name", "class_id", "hour_total", "hour_remark", "hour_week", "field"],
     ({"name": "Calculus", "week": [range(1, 17)], "day": 1, "time": range(1, 3)},
      "<ScheduleCourse Calculus week=[range(1, 17)] day=1 time=range(1, 3)>"))
])
def test_dict_model(model, members, repr_pair):
    kwargs = {member: idx for idx, member in zip(range(len(members)), members)}
    model_1 = model(**kwargs)
    for idx, member in zip(range(len(members)), members):
        assert getattr(model_1, member) == idx

    with pytest.raises(TypeError):
        model(fake_arg=0)

    repr_kwargs = repr_pair[0]
    repr_kwargs.update({member: None for member in members if member not in repr_kwargs})
    assert repr(model(**repr_kwargs)) == repr_pair[1]


def test_selection_sector(mocker):
    hash_func = mocker.patch("pysjtu.models.selection.elfhash", return_value=919293949)
    class_func = mocker.Mock(return_value=["1"])

    members = ["task_type", "xkly", "pe_op_param", "sector_type_id", "txbsfrl", "kkbk", "course_type_code", "name",
               "xkkz_id", "shared_info", "include_other_grades", "include_other_majors", "sfznkx", "zdkxms"]
    kwargs = {member: idx for idx, member in zip(range(len(members)), members) if member != "name"}
    model_1 = SelectionSector(name="dummy", _func_classes=class_func, **kwargs)

    assert hash(model_1) == 919293949
    hash_func.assert_called_once()
    hash_func.reset_mock()
    assert hash(model_1) == 919293949
    hash_func.assert_not_called()

    assert model_1.classes == ["1"]
    class_func.assert_called_once()


def test_score_detail(mocker):
    members = ["name", "teacher", "score", "credit", "gp", "invalid", "course_type", "category", "score_type", "method",
               "course_id", "class_name", "class_id"]
    kwargs = {member: idx for idx, member in zip(range(len(members)), members) if member != "class_id"}
    model_1 = Score(**kwargs)
    model_1.year = 2012
    model_1.term = 1
    model_1.class_id = "dummy"
    fake_detail_func = mocker.Mock(return_value="fake detail")
    model_1._func_detail = fake_detail_func
    for _ in range(2):
        assert model_1.detail == "fake detail"
        fake_detail_func.assert_called_once_with(2012, 1, "dummy")


@pytest.fixture
def fake_model():
    class FakeModel:
        def __init__(self, **kwargs):
            self.__dict__.update(**kwargs)

        def __eq__(self, other):
            return self.__dict__ == other.__dict__

        def __repr__(self):
            return f"<FakeModel {self.__dict__}>"

    return FakeModel


@pytest.mark.parametrize("model, mock_schema, test_score",
                         [(Exams, ExamSchema, False), (Schedule, ScheduleCourseSchema, False),
                          (Scores, ScoreSchema, True)])
def test_loader_model(mocker, fake_model, model, mock_schema, test_score):
    rtn_var = [
        fake_model(name="Calculus", credit=6.0),
        fake_model(name="Calculus", credit=4.0),
        fake_model(name="Chemistry", credit=4.0)
    ]
    if test_score:
        loaded_var = rtn_var.copy()
        for item in loaded_var:
            item.__dict__.update({"year": 2012, "term": 1, "_func_detail": None})
    else:
        loaded_var = rtn_var
    mocker.patch.object(mock_schema, "load", return_value=rtn_var)
    if test_score:
        model_1 = model(year=2012, term=1, func_detail=None)
    else:
        model_1 = model(year=2012, term=1)
    assert model_1.year == 2012
    assert model_1.term == 1
    model_1.load(None)
    assert len(model_1) == 3
    assert model_1 == rtn_var
    assert model_1.filter(name="Calculus") == loaded_var[:2]
    assert model_1.filter(name="Calculus", credit=4.0) == [loaded_var[1]]
    with pytest.raises(KeyError):
        model_1.filter(dummy="Lorem")


def test_schedule_filter(mocker, fake_model):
    rtn_var = [
        fake_model(name="Calculus", day=1, week=[range(1, 17)], time=range(1, 3)),
        fake_model(name="Chemistry", day=3, week=[range(1, 14, 2)], time=range(1, 3)),
        fake_model(name="Millitary", day=2, week=[5, 10, range(14, 16)], time=range(5, 7))
    ]
    mocker.patch.object(ScheduleCourseSchema, "load", return_value=rtn_var)
    model_1 = Schedule()
    model_1.load(None)
    assert model_1.filter(week=3) == rtn_var[:2]
    assert model_1.filter(week=17) == []
    assert model_1.filter(week=[5]) == rtn_var
    assert model_1.filter(week=range(15, 17)) == [rtn_var[0], rtn_var[2]]
    assert model_1.filter(week=[3, range(16, 17)]) == rtn_var[:2]
    assert model_1.filter(week=[5, range(16, 17)]) == rtn_var

    assert model_1.filter(day=2) == [rtn_var[-1]]
    assert model_1.filter(day=range(1, 5, 2)) == rtn_var[:2]
    assert model_1.filter(day=[2, range(1, 5, 2)]) == rtn_var

    assert model_1.filter(time=2) == rtn_var[:2]
    assert model_1.filter(time=3) == []
    assert model_1.filter(time=[6]) == [rtn_var[-1]]
    assert model_1.filter(time=range(4, 7)) == [rtn_var[-1]]
    assert model_1.filter(time=[1, range(4, 7)]) == rtn_var



import time
from typing import Union

from httpx.config import (TimeoutTypes, UNSET, UnsetType)

from pysjtu import consts
from pysjtu import models
from pysjtu import schemas
from pysjtu.client.base import BaseClient
from pysjtu.exceptions import GPACalculationException


class GPAMixin(BaseClient):
    _default_gpa_query_params: models.GPAQueryParams

    def __init__(self):
        super().__init__()
        # noinspection PyTypeChecker
        self._default_gpa_query_params = None

    @property
    def default_gpa_query_params(self) -> models.GPAQueryParams:
        """ Get default gpa query params defined by the website. """
        if not self._default_gpa_query_params:
            rtn = self._session.get(consts.GPA_PARAMS_URL,
                                    params={"_": int(time.time() * 1000), "su": self.student_id})
            self._default_gpa_query_params = schemas.GPAQueryParamsSchema().load(rtn.json())  # type: ignore

        return self._default_gpa_query_params

    def gpa(self, query_params: models.GPAQueryParams, timeout: Union[TimeoutTypes, UnsetType] = UNSET) -> models.GPA:
        """
        Query your GP & GPA and their rankings of specific year & term.

        :param query_params: parameters for this query.
            A default one can be fetched by reading property `default_gpa_query_params`.
        :param timeout: (optional) How long to wait for the server to send data before giving up.
        :return: A new :class:`GPA` object.
        """
        compiled_params = schemas.GPAQueryParamsSchema().dump(query_params)
        calc_rtn = self._session.post(consts.GPA_CALC_URL + str(self.student_id),
                                      data=compiled_params, timeout=timeout)
        if calc_rtn.text != "\"统计成功！\"":
            if calc_rtn.text == "\"统计失败！\"":
                raise GPACalculationException("Calculation failure.")
            if "无功能权限" in calc_rtn.text:
                raise GPACalculationException("Unauthorized.")
        compiled_params.update({"_search": False,
                                "nd": int(time.time() * 1000), "queryModel.showCount": 15,
                                "queryModel.currentPage": 1, "queryModel.sortName": "",
                                "queryModel.sortOrder": "asc", "time": 0})
        raw = self._session.post(consts.GPA_QUERY_URL + str(self.student_id),
                                 data=compiled_params, timeout=timeout)
        return schemas.GPASchema().load(raw.json()["items"][0])  # type: ignore



from datetime import datetime

from pysjtu.models.base import Gender
from pysjtu.parser.profile import ProfileField

# @formatter:off
profile_fields = [
    ProfileField("student_id",              "/html/body/div[1]/div/div/form/div/div[2]/div/div/div/div/div[1]/div/div[1]/div/div[1]/div/div/p", int),
    ProfileField("name",                    "/html/body/div[1]/div/div/form/div/div[2]/div/div/div/div/div[1]/div/div[1]/div/div[2]/div/div/p"),
    ProfileField("name_pinyin",             "/html/body/div[1]/div/div/form/div/div[2]/div/div/div/div/div[1]/div/div[1]/div/div[3]/div/div/p"),
    ProfileField("former_name",             "/html/body/div[1]/div/div/form/div/div[2]/div/div/div/div/div[1]/div/div[1]/div/div[4]/div/div/p"),
    ProfileField("gender",                  "/html/body/div[1]/div/div/form/div/div[2]/div/div/div/div/div[1]/div/div[1]/div/div[5]/div/div/p", lambda x: Gender.male if x == "男" else Gender.female),
    ProfileField("certificate_type",        "/html/body/div[1]/div/div/form/div/div[2]/div/div/div/div/div[1]/div/div[1]/div/div[6]/div/div/p"),
    ProfileField("certificate_number",      "/html/body/div[1]/div/div/form/div/div[2]/div/div/div/div/div[1]/div/div[1]/div/div[7]/div/div/p", int),
    ProfileField("birth_date",              "/html/body/div[1]/div/div/form/div/div[2]/div/div/div/div/div[1]/div/div[4]/div/div/p", lambda x: datetime.strptime(x, "%Y-%m-%d").date()),
    ProfileField("enrollment_date",         "/html/body/div[1]/div/div/form/div/div[2]/div/div/div/div/div[1]/div/div[8]/div/div/p", lambda x: datetime.strptime(x, "%Y-%m-%d").date()),
    ProfileField("birthplace",              "/html/body/div[1]/div/div/form/div/div[2]/div/div/div/div/div[1]/div/div[12]/div/div/p"),
    ProfileField("ethnicity",               "/html/body/div[1]/div/div/form/div/div[2]/div/div/div/div/div[1]/div/div[5]/div/div/p"),
    ProfileField("native_place",            "/html/body/div[1]/div/div/form/div/div[2]/div/div/div/div/div[1]/div/div[9]/div/div/p"),
    ProfileField("foreign_status",          "/html/body/div[1]/div/div/form/div/div[2]/div/div/div/div/div[1]/div/div[13]/div/div/p"),
    ProfileField("political_status",        "/html/body/div[1]/div/div/form/div/div[2]/div/div/div/div/div[1]/div/div[7]/div/div/p"),
    ProfileField("enrollment_province",     "/html/body/div[1]/div/div/form/div/div[2]/div/div/div/div/div[1]/div/div[11]/div/div/p"),
    ProfileField("nationality",             "/html/body/div[1]/div/div/form/div/div[2]/div/div/div/div/div[1]/div/div[14]/div/div/p"),
    ProfileField("domicile_place",          "/html/body/div[1]/div/div/form/div/div[2]/div/div/div/div/div[1]/div/div[10]/div/div/p"),
    ProfileField("cee_candidate_number",    "/html/body/div[1]/div/div/form/div/div[2]/div/div/div/div/div[3]/div/div[1]/div/div/p", int),
    ProfileField("middle_school",           "/html/body/div[1]/div/div/form/div/div[2]/div/div/div/div/div[3]/div/div[2]/div/div/p"),
    ProfileField("religion",                "/html/body/div[1]/div/div/form/div/div[2]/div/div/div/div/div[3]/div/div[5]/div/div/p"),
    ProfileField("email",                   "/html/body/div[1]/div/div/form/div/div[2]/div/div/div/div/div[4]/div/div[1]/div/div/p"),
    ProfileField("cellphone",               "/html/body/div[1]/div/div/form/div/div[2]/div/div/div/div/div[4]/div/div[2]/div/div/p", int),
    ProfileField("family_address",          "/html/body/div[1]/div/div/form/div/div[2]/div/div/div/div/div[4]/div/div[3]/div/div/p"),
    ProfileField("mailing_address",         "/html/body/div[1]/div/div/form/div/div[2]/div/div/div/div/div[4]/div/div[4]/div/div/p"),
    ProfileField("landline",                "/html/body/div[1]/div/div/form/div/div[2]/div/div/div/div/div[4]/div/div[5]/div/div/p", int),
    ProfileField("zip_code",                "/html/body/div[1]/div/div/form/div/div[2]/div/div/div/div/div[4]/div/div[6]/div/div/p", int)
]
# @formatter:on



from datetime import date
from os import path

import pytest
from lxml import html

from pysjtu.models.base import Gender
from pysjtu.parser.profile import ProfileField, parse
from pysjtu.parser.selection import parse_fields, parse_sectors
from pysjtu.schemas.profile import profile_fields


@pytest.fixture()
def website_loader():
    RESP_DIR = path.join(path.dirname(path.abspath(__file__)), 'resources/website')

    def _resp_loader(name):
        with open(path.join(RESP_DIR, name + ".html"), encoding="utf-8") as f:
            return f.read()

    return _resp_loader


def test_parse_sectors(website_loader):
    raw_src = website_loader("zzxkyzb_cxZzxkYzbIndex")
    assert parse_sectors(raw_src) == [
        ("01", "A000000000000B76E055F8163ED16360", "主修课程"),
        ("71", "A000000000000A7FE055F8163ED16360", "民族生课程"),
        ("72", "A000000000000881E055F8163ED16360", "留学生课程"),
        ("06", "A000000000000F12E055F8163ED16360", "板块课(体育（3）)"),
        ("10", "A000000000000455E055F8163ED16360", "通识课"),
        ("11", "A000000000000B28E055F8163ED16360", "通选课")
    ]


def test_parse_fields(website_loader):
    raw_src = website_loader("zzxkyzb_cxZzxkYzbIndex")
    assert parse_fields(raw_src, ["xqh_id", "zyh_id", "njdm_id"]) == \
           {"xqh_id": "02", "zyh_id": "000000", "njdm_id": "2019"}
    with pytest.raises(StopIteration):
        parse_fields(raw_src, ["xqh_id", "invalid_field"])


def test_parse_profile_single():
    el = html.fromstring(
        '<!doctype html> <html> <head> <title>LightQuantum</title> </head> <body> '
        '<p>test<a href="TEST">TEST FIELD</a>.</p> </body> </html>')
    el2 = html.fromstring(
        '<!doctype html> <html> <head> <title>LightQuantum</title> </head> <body> '
        '<p>test<a href="TEST"></a>.</p> </body> </html>')
    el3 = html.fromstring(
        '<!doctype html> <html> <head> <title>LightQuantum</title> </head> <body> '
        '<p>test.</p> </body> </html>')
    field = ProfileField("test_field", "/html/body/p/a", lambda x: x.replace("FIELD", "field"))
    assert field.parse(el) == "TEST field"
    assert field.parse(el2) is None
    assert field.parse(el3) is None


def test_parse_profile(website_loader):
    raw_src = website_loader("xsgrxxwh_cxXsgrxx")
    profile = parse(profile_fields, raw_src)
    assert profile == {'student_id': 519027910001, 'name': '林芳', 'name_pinyin': None, 'former_name': None,
                       'gender': Gender.female, 'certificate_type': '居民身份证', 'certificate_number': 300000000000000000,
                       'birth_date': date(2000, 11, 1), 'enrollment_date': date(2019, 9, 7), 'birthplace': None,
                       'ethnicity': '人族', 'native_place': 'OPUS_H', 'foreign_status': None, 'political_status': '地球教女巫',
                       'enrollment_province': 'OPUS_O', 'nationality': 'OPUS_N', 'domicile_place': None,
                       'cee_candidate_number': 19300000000000, 'middle_school': '地球教教会学校', 'religion': None,
                       'email': 'linfei@sjtu.edu.cn', 'cellphone': 17000000000, 'family_address': '奥伯斯火箭工厂_H',
                       'mailing_address': '奥伯斯火箭工厂_M', 'landline': 10000000, 'zip_code': 200000}



from functools import partial
from typing import Union

from httpx.config import (TimeoutTypes, UNSET, UnsetType)

from pysjtu import consts
from pysjtu import models
from pysjtu import schemas
from pysjtu.client.base import BaseClient
from pysjtu.utils import range_list_to_str, schema_post_loader


class CourseLibMixin(BaseClient):
    def __init__(self):
        super().__init__()

    def query_courses(self, year: int, term: int, page_size: int = 15, name: str = None, teacher: str = None,
                      day_of_week: list = None, week: list = None, time_of_day: list = None,
                      timeout: Union[TimeoutTypes, UnsetType] = UNSET) -> models.QueryResult[models.LibCourse]:
        """
        Query courses matching given criteria from the whole course lib of SJTU.

        :param year: year in which target courses are given.
        :param term: term in which target courses are given.
        :param page_size: page size for result iteration.
        :param name: (optional) Name (can be fuzzy) of target courses.
        :param teacher: (optional) Teacher name of target courses.
        :param day_of_week: (optional) Day of week of target courses.
        :param week: (optional) Week of target courses.
        :param time_of_day: (optional) Time of day of target courses.
        :param timeout: (optional) How long to wait for the server to send data before giving up.
        :return: A new :class:`QueryResult` object.
        """
        _args = {"year": "xnm", "term": "xqm", "name": "kch_id", "teacher": "jqh_id", "day_of_week": "xqj",
                 "week": "qsjsz", "time_of_day": "skjc"}
        year = year
        term = consts.TERMS[term]
        name = name if name else ''
        teacher = teacher if teacher else ''
        day_of_week = range_list_to_str(day_of_week) if day_of_week else []
        week = range_list_to_str(week) if week else []
        time_of_day = range_list_to_str(time_of_day) if time_of_day else []
        req_params = {}
        for (k, v) in _args.items():
            if k in dir():
                req_params[v] = locals()[k]

        req = partial(self._session.post, consts.COURSELIB_URL + str(self.student_id), timeout=timeout)

        return models.QueryResult(req, partial(schema_post_loader, schemas.LibCourseSchema), req_params,
                                  page_size=page_size)



import re
import typing
from typing import List

from marshmallow import EXCLUDE, Schema, fields, post_load  # type: ignore

from pysjtu.consts import CHINESE_WEEK
from pysjtu.models.selection import Gender, LessonTime
from pysjtu.schemas.base import SplitField, StrBool
from pysjtu.utils import parse_course_week


class GenderField(fields.Field):
    def _deserialize(
            self,
            value: typing.Any,
            attr: typing.Optional[str],
            data: typing.Optional[typing.Mapping[str, typing.Any]],
            **kwargs
    ):
        if not value:
            return None  # pragma: no cover
        return Gender(int(value))

    def _serialize(self, value: typing.Any, attr: str, obj: typing.Any, **kwargs):
        return value.value


class TimeField(fields.Field):
    regex = re.compile("星期(?P<weekday>.)第(?P<time>.*?)节{(?P<week>.*?)}")

    def _deserialize(
            self,
            value: typing.Any,
            attr: typing.Optional[str],
            data: typing.Optional[typing.Mapping[str, typing.Any]],
            **kwargs
    ):
        if not value:
            return None  # pragma: no cover

        def _parse_time(input_str: str) -> List[range]:
            return [range(int(time[0]), int(time[1]) + 1)
                    for time in [times.split('-') for times in input_str.split(",")]]

        def _dict_to_time(input_dict: dict) -> LessonTime:
            return LessonTime(
                weekday=CHINESE_WEEK[input_dict["weekday"]],
                week=parse_course_week(input_dict["week"]),
                time=_parse_time(input_dict["time"])
            )

        return [_dict_to_time(match.groupdict()) for match in self.regex.finditer(value)]


class TeacherField(fields.Field):
    def _deserialize(
            self,
            value: typing.Any,
            attr: typing.Optional[str],
            data: typing.Optional[typing.Mapping[str, typing.Any]],
            **kwargs
    ):
        if not value:
            return None  # pragma: no cover
        return [tuple(teacher.split("/")[1:]) for teacher in value.split(";")]


class SelectionSharedInfoSchema(Schema):
    term = fields.Str(required=True, data_key="xqh_id")
    selection_year = fields.Int(required=True, data_key="xkxnm")
    selection_term = fields.Int(required=True, data_key="xkxqm")
    major_id = fields.Str(required=True, data_key="zyh_id")
    student_grade = fields.Int(required=True, data_key="njdm_id")
    natural_class_id = fields.Str(required=True, data_key="bh_id")
    self_selecting_status = fields.Int(required=True, data_key="xszxzt")
    ccdm = fields.Str(required=True)
    student_type_code = fields.Int(required=True, data_key="xslbdm")
    gender = GenderField(required=True, data_key="xbm")
    field_id = fields.Str(required=True, data_key="zyfx_id")
    student_background = fields.Int(required=True, data_key="xsbj")

    # noinspection PyUnusedLocal
    @post_load
    def wrap(self, data, **kwargs):
        return SelectionSharedInfo(**data)


class SelectionSectorSchema(Schema):
    task_type = fields.Int(required=True, data_key="rwlx")
    xkly = fields.Int(required=True)
    pe_op_param = fields.Int(required=True, data_key="tykczgxdcs")
    sector_type_id = fields.Str(required=True, data_key="bklx_id")
    course_type_code = fields.Str(required=True, dump_only=True, data_key="kklxdm")
    include_other_grades = StrBool(required=True, data_key="sfkknj")
    include_other_majors = StrBool(required=True, data_key="sfkkzy")
    sfznkx = StrBool(required=True)
    zdkxms = fields.Int(required=True)
    txbsfrl = fields.Int(required=True)
    kkbk = fields.Int(required=True)
    xkkz_id = fields.Str(dump_only=True)

    # noinspection PyUnusedLocal
    @post_load
    def wrap(self, data, **kwargs):
        return SelectionSector(**data)


class SelectionCourseSchema(Schema):
    class Meta:
        unknown = EXCLUDE

    name = fields.Str(required=True, data_key="kcmc")
    credit = fields.Float(required=True, data_key="xf")
    course_id = fields.Str(required=True, data_key="kch")
    internal_course_id = fields.Str(required=True, data_key="kch_id")
    class_name = fields.Str(required=True, data_key="jxbmc")
    class_id = fields.Str(required=True, data_key="jxb_id")
    students_registered = fields.Int(required=True, data_key="yxzrs")


class SelectionClassSchema(Schema):
    class Meta:
        unknown = EXCLUDE

    class_id = fields.Str(required=True, data_key="jxb_id")
    register_id = fields.Str(required=True, data_key="do_jxb_id")
    teachers = TeacherField(required=True, data_key="jsxx")
    locations = SplitField(required=True, data_key="jxdd", sep="<br/>")
    time = TimeField(required=True, data_key="sksj")
    course_type = SplitField(required=True, data_key="kcxzmc", sep=",")
    remark = fields.Str(data_key="xkbz", missing=None)
    students_planned = fields.Int(required=True, data_key="jxbrl")


from pysjtu.models.selection import SelectionSharedInfo, SelectionSector



import typing

from marshmallow import ValidationError, fields  # type: ignore

from pysjtu.utils import parse_course_week


class StrBool(fields.Field):
    def _deserialize(
            self,
            value: typing.Any,
            attr: typing.Optional[str],
            data: typing.Optional[typing.Mapping[str, typing.Any]],
            **kwargs
    ):
        if not value:
            return None  # pragma: no cover
        if value in ["0"]:
            return False
        if value in ["1"]:
            return True
        raise ValidationError("StrBool values must be one of [0, 1].")

    def _serialize(self, value: typing.Any, attr: str, obj: typing.Any, **kwargs):
        if value is None:
            return ""  # pragma: no cover
        if value is False:
            return "0"
        if value is True:
            return "1"
        raise ValidationError("Invalid bool value.")


class ChineseBool(fields.Field):
    def _deserialize(
            self,
            value: typing.Any,
            attr: typing.Optional[str],
            data: typing.Optional[typing.Mapping[str, typing.Any]],
            **kwargs
    ):
        if not value:
            return None  # pragma: no cover
        return value == "是"


class SplitField(fields.Field):
    def __init__(self, sep: str, *args, **kwargs):
        self.sep = sep
        super().__init__(*args, **kwargs)

    def _deserialize(
            self,
            value: typing.Any,
            attr: typing.Optional[str],
            data: typing.Optional[typing.Mapping[str, typing.Any]],
            **kwargs
    ):
        if not value:
            return None  # pragma: no cover
        return value.split(self.sep)

    def _serialize(self, value: typing.Any, attr: str, obj: typing.Any, **kwargs):
        return self.sep.join(str(item) for item in value)


class CourseTime(fields.Field):
    def _deserialize(
            self,
            value: typing.Any,
            attr: typing.Optional[str],
            data: typing.Optional[typing.Mapping[str, typing.Any]],
            **kwargs
    ):
        if not value:
            return None  # pragma: no cover
        value = value.replace("节", "")
        cs = list(map(int, value.split("-")))
        return list(cs) if len(cs) == 1 else range(cs[0], cs[1] + 1)


class CourseWeek(fields.Field):
    def _deserialize(
            self,
            value: typing.Any,
            attr: typing.Optional[str],
            data: typing.Optional[typing.Mapping[str, typing.Any]],
            **kwargs
    ):
        if not value:
            return None  # pragma: no cover
        return parse_course_week(value)



import re
from datetime import date, datetime

from pysjtu import consts
from pysjtu.client.api import CourseLibMixin, ExamMixin, GPAMixin, ScheduleMixin, ScoreMixin, SelectionMixin, ProfileMixin
from pysjtu.client.base import BaseClient
from pysjtu.session import BaseSession, Session
from pysjtu.ocr import JCSSRecognizer


class Client(ProfileMixin, SelectionMixin, ScheduleMixin, CourseLibMixin, ExamMixin, GPAMixin, ScoreMixin, BaseClient):
    """
    A pysjtu client with schedule query, score query, exam query, etc.

    Usage::

        >>> import pysjtu
        >>> s = pysjtu.Session(username="user@sjtu.edu.cn", password="something_secret")
        >>> client = pysjtu.Client(session=s)
        >>> sched = client.schedule(2019, 0)
        >>> sched
        [<ScheduleCourse 军事理论 week=[range(9, 17)] day=1 time=range(1, 3)>, ...]
        >>> sched.filter(time=range(3,5), day=range(2, 4))
        [<ScheduleCourse 程序设计思想与方法（C++） week=[range(1, 10), range(11, 17)] day=2 time=range(3, 5)>,
        <ScheduleCourse 大学英语（4） week=[range(1, 17)] day=3 time=range(3, 5)>]

    :param session: The :class:`Session` to be built upon.
    """
    _session: BaseSession
    _term_start: date

    def __init__(self, session: BaseSession):
        super().__init__()
        if not isinstance(session, BaseSession):
            raise TypeError("'session' isn't an instance of BaseSession.")
        self._session = session

        # noinspection PyTypeChecker
        self._term_start = None  # type: ignore

    @property
    def term_start_date(self) -> date:
        """ Get the term start date for the current term. """
        if not self._term_start:
            raw = self._session.get(consts.CALENDAR_URL + str(self.student_id))
            self._term_start = datetime.strptime(min(re.findall(r"\d{4}-\d{2}-\d{2}", raw.text)), "%Y-%m-%d").date()
        return self._term_start

    # noinspection PyProtectedMember
    @property
    def student_id(self) -> int:
        """ Get the student id of the current session. """
        if "student_id" not in self._session._cache_store:
            rtn = self._session.get(consts.HOME_URL)
            self._session._cache_store["student_id"] = int(
                re.findall(r"(?<=id=\"sessionUserKey\" value=\")\d*", rtn.text)[0])
        return self._session._cache_store["student_id"]


def create_client(username: str, password: str, use_jcss: bool = True, _mocker_app=None) -> Client:
    """
    Create a new :class:`Client` with default options.
    To change :class:`Session` settings or preserve your session, use :class:`Session` and :class:`Client` instead.

    :param username: JAccount username.
    :param password: JAccount password.
    :param use_jcss: Use JAccount Captcha Solver Service to recognize captcha instead of built-in ONNX based recognizer.
    :param _mocker_app: An WSGI application to send requests to (for debug or test purposes).
    :return: an authenticated :class:`Client`.
    """
    sess = Session(username=username, password=password, _mocker_app=_mocker_app, ocr=JCSSRecognizer() if use_jcss else None)
    return Client(session=sess)



from .base import LazyResult, PARTIAL, QueryResult, Result, Results
from .course import LibCourse
from .exam import Exam, Exams
from .gpa import CourseRange, GPA, GPAQueryParams, LogicEnum, Ranking
from .profile import Profile
from .schedule import Schedule, ScheduleCourse
from .score import Score, ScoreFactor, Scores
from .selection import SelectionClass, SelectionSector, SelectionSharedInfo



import base64
import collections
from math import inf
from pathlib import Path
from typing import BinaryIO, Union

FileTypes = Union[BinaryIO, str, Path]


def elfhash(input_str: str):
    s = base64.b64encode(input_str.encode("utf-8"))
    _hash = 0
    x = 0
    for c in s:
        _hash = (_hash << 4) + c
        x = _hash & 0xF0000000
        if x:
            _hash ^= (x >> 24)
            _hash &= ~x
    return _hash & 0x7FFFFFFF


def parse_course_week(value):
    def _parse(item):
        if item[-2] in ["单", "双"]:
            start, end = map(int, item[:-4].split('-'))
            start += (1 - start % 2) if item[-2] == "单" else (start % 2)
            return range(start, end + 1, 2)
        else:
            x = list(map(int, item[:-1].split('-')))
            return x[0] if len(x) == 1 else range(x[0], x[1] + 1)  # type: ignore

    return [_parse(item) for item in value.split(",")]


def parse_slice(val):
    if isinstance(val, int):
        return val
    if hasattr(val, "__index__") and isinstance(val.__index__, collections.abc.Callable):
        return val.__index__()
    if val is None:
        return val
    raise AttributeError("slice indices must be integers or None or have an __index__ method")


def has_callable(obj, name):
    return callable(getattr(obj, name, None))


def replace_keys(data, pairs):
    for from_key, to_key in pairs:
        if from_key in data:
            data[to_key] = data.pop(from_key)
    return data


def schema_post_loader(schema_ref, data):
    if isinstance(data, list):
        return schema_ref(many=True).load(data)
    if isinstance(data, dict):
        return schema_ref().load(data)
    raise TypeError


def range_list_to_str(range_list):
    return "或".join((str(x) for x in flatten(range_list)))


# skipcq: PYL-R1710
def range_in_set(set_in):
    if len(set_in) == 0:
        return set()
    last_elem = -inf
    start = None
    for elem in set_in:
        if elem != last_elem + 1:
            if start:
                yield range(start, last_elem + 1)
            start = elem
        last_elem = elem
    yield range(start, last_elem + 1)


def overlap(list1, list2):
    if not isinstance(list1, list):
        list1 = [list1]
    if not isinstance(list2, list):
        list2 = [list2]
    flatten1 = flatten(list1)
    flatten2 = flatten(list2)
    set1 = set(flatten1)
    return set1.intersection(flatten2)


def flatten(obj):
    for el in obj:
        if isinstance(el, collections.abc.Iterable) and not isinstance(el, (str, bytes)):
            yield from flatten(el)
        else:
            yield el



import time
from functools import partial
from typing import List, Union

from httpx.config import (TimeoutTypes, UNSET, UnsetType)

from pysjtu import consts
from pysjtu import models
from pysjtu import schemas
from pysjtu.client.base import BaseClient


class ScoreMixin(BaseClient):
    def __init__(self):
        super().__init__()

    def _get_score_detail(self, year: int, term: int, class_id: str, timeout: Union[TimeoutTypes, UnsetType] = UNSET) \
            -> List[models.ScoreFactor]:
        raw = self._session.post(consts.SCORE_DETAIL_URL + str(self.student_id),
                                 data={"xnm": year, "xqm": consts.TERMS[term], "jxb_id": class_id, "_search": False,
                                       "nd": int(time.time() * 1000), "queryModel.showCount": 15,
                                       "queryModel.currentPage": 1, "queryModel.sortName": "",
                                       "queryModel.sortOrder": "asc", "time": 1}, timeout=timeout)
        factors = schemas.ScoreFactorSchema(many=True).load(raw.json()["items"][:-1])  # type: ignore
        return factors

    def score(self, year: int, term: int, timeout: Union[TimeoutTypes, UnsetType] = UNSET) \
            -> models.Results[models.Score]:
        """
        Fetch your scores of specific year & term.

        :param year: year for the new :class:`Scores` object.
        :param term: term for the new :class:`Scores` object.
        :param timeout: (optional) How long to wait for the server to send data before giving up.
        :return: A new :class:`Scores` object.
        :rtype: :class:`Scores`
        """
        raw = self._session.post(consts.SCORE_URL,
                                 data={"xnm": year, "xqm": consts.TERMS[term], "_search": False,
                                       "nd": int(time.time() * 1000), "queryModel.showCount": 15,
                                       "queryModel.currentPage": 1, "queryModel.sortName": "",
                                       "queryModel.sortOrder": "asc", "time": 1}, timeout=timeout)
        scores = models.Scores(year, term, partial(self._get_score_detail, timeout=timeout))
        scores.load(raw.json()["items"])  # type: ignore
        return scores



from dataclasses import dataclass
from typing import Callable, List, Optional

from pysjtu.models.base import Result, Results


@dataclass
class ScoreFactor(Result):
    """
    A model which describes detailed composition of a course's score.

    :param name: item name
    :param percentage: item factor
    :param score: item score
    """
    name: str
    percentage: float
    score: float

    def __repr__(self):
        return f"<ScoreFactor {self.name}({self.percentage * 100}%)={self.score}>"


@dataclass
class Score(Result):
    """
    A model which describes the score of a specific course. Some fields may be empty.

    :param name: literal name of the course.
    :type name: str
    :param teacher: the teacher who offers this course.
    :type teacher: str
    :param score: score of this course
    :type score: str
    :param credit: credits that the course provides.
    :type credit: float
    :param gp: gp earned in this course.
    :type gp: float
    :param invalid: whether this score is voided.
    :type invalid: bool
    :param detail: a ScoreFactor object representing detailed composition of the score.
    :type detail: List[:class:`ScoreFactor`]
    :param course_type: type of this course. (compulsory, elective, etc)
    :type course_type: str
    :param category: category of this course. (specialized, general, PE, etc)
    :type category: str
    :param score_type: type of your score. (acquired by normal examination, etc)
    :type score_type: str
    :param method: assessment method of this course. (exams, assesses, etc)
    :type method: str
    :param course_id: course id.
    :type course_id: str
    :param class_name: class name (constant between years).
    :type class_name: str
    :param class_id: class id (variable between years).
    :type class_id: str
    """
    name: str
    teacher: str
    score: str
    credit: float
    gp: float
    invalid: Optional[bool] = None
    course_type: Optional[str] = None
    category: Optional[str] = None
    score_type: Optional[str] = None
    method: Optional[str] = None
    course_id: Optional[str] = None
    class_name: Optional[str] = None
    class_id: Optional[str] = None
    year: int = 0
    term: int = 0
    _detail: List[ScoreFactor] = None
    _func_detail: Callable = None

    def __repr__(self):
        return f"<Score {self.name} score={self.score} credit={self.credit} gp={self.gp}>"

    @property
    def detail(self) -> List[ScoreFactor]:
        if not self._detail:
            self._detail = self._func_detail(self.year, self.term, self.class_id)
        return self._detail


from pysjtu.schemas.score import ScoreSchema


class Scores(Results[Score]):
    """
    A list-like interface to Score collections.
    An additional filter method has been added to make filter operations easier.
    """
    _schema = ScoreSchema
    _result_model = Score

    def __init__(self, year: int = 0, term: int = 0, func_detail: Callable = None):
        super().__init__(year, term)
        self._func_detail = func_detail

    def load(self, data: dict):
        """
        Load a list of dicts into Scores, and deserialize dicts to Score objects.

        :param data: a list of dicts contains scores.
        """
        super().load(data)
        for item in self:
            item.year = self.year
            item.term = self.term
            # noinspection PyTypeHints
            item._func_detail = self._func_detail  # type: ignore



from .course import LibCourseSchema
from .exam import ExamSchema
from .gpa import GPAQueryParamsSchema, GPASchema
from .schedule import ScheduleCourseSchema
from .score import ScoreFactorSchema, ScoreSchema
from .selection import SelectionClassSchema, SelectionCourseSchema, SelectionSectorSchema, SelectionSharedInfoSchema



import typing

from marshmallow import EXCLUDE, Schema, fields, post_load  # type: ignore

from pysjtu.schemas.base import CourseTime, CourseWeek, SplitField


class LibCreditHourDetail(fields.Field):
    def _deserialize(
            self,
            value: typing.Any,
            attr: typing.Optional[str],
            data: typing.Optional[typing.Mapping[str, typing.Any]],
            **kwargs
    ):
        if not value:
            return None  # pragma: no cover
        class_hour_details = value.split("-")
        rtn = {}
        for item in class_hour_details:
            name, hour = item.split("(")
            rtn[name] = float(hour[:-1])
        return rtn


class LibCourseSchema(Schema):
    class Meta:
        unknown = EXCLUDE

    name = fields.Str(required=True, data_key="kcmc")
    day = fields.Int(data_key="xqj")
    week = CourseWeek(data_key="qsjsz")
    time = CourseTime(data_key="skjc")
    location = fields.Str(data_key="cdmc")
    locations = SplitField(data_key="jxdd", sep=";")
    faculty = fields.Str(data_key="kkxy")
    credit = fields.Float(data_key="xf")
    teacher = SplitField(data_key="zjs", sep=",")
    course_id = fields.Str(data_key="kch")
    class_name = fields.Str(data_key="jxbmc")
    class_id = fields.Str(data_key="jxb_id")
    class_composition = SplitField(data_key="jxbzc", sep=";")
    hour_total = fields.Int(data_key="rwzxs")
    hour_remark = LibCreditHourDetail(data_key="zhxs")
    seats = fields.Int(data_key="zws")
    students_elected = fields.Int(data_key="xkrs")
    students_planned = fields.Int(data_key="jxbrs")

    # noinspection PyUnusedLocal
    @post_load
    def wrap(self, data, **kwargs):
        return LibCourse(**data)


from pysjtu.models.course import LibCourse



import io
import pickle
import re
import time
import warnings
from functools import partial
from http.cookiejar import CookieJar
from pathlib import Path
from typing import Any, Callable, Dict, Optional, Union
from urllib.parse import parse_qs, urljoin, urlparse

import httpx
from httpx.auth import AuthTypes
from httpx.config import (ProxiesTypes, TimeoutTypes, UNSET, UnsetType)
from httpx.dispatch.base import SyncDispatcher
from httpx.models import (
    CookieTypes,
    HeaderTypes,
    QueryParamTypes,
    RequestData,
    RequestFiles,
    Response,
    URLTypes,
)

from pysjtu.ocr import JCSSRecognizer, Recognizer
from . import consts
from .exceptions import DumpWarning, LoadWarning, LoginException, ServiceUnavailable, SessionException
from .utils import FileTypes


class BaseSession:
    """ Base session """
    _cache_store: dict

    def get(self, *args, **kwargs):
        raise NotImplementedError  # pragma: no cover

    def post(self, *args, **kwargs):
        raise NotImplementedError  # pragma: no cover


class Session(BaseSession):
    """
    A pysjtu session with login management, cookie persistence, etc.

    Usage::

        >>> import pysjtu
        >>> s = pysjtu.Session()
        >>> s.login('user@sjtu.edu.cn', 'something_secret')
        >>> s.get('https://i.sjtu.edu.cn')
        <Response [200 OK]>
        >>> s.dump('session_file')

    Or as a context manager::

        >>> with pysjtu.Session(username='user@sjtu.edu.cn', password='something_secret') as s:
        ...     s.get('https://i.sjtu.edu.cn')
        ...     s.dump('session_file')
        <Response [200 OK]>

        >>> with pysjtu.Session(session_file='session_file', mode='r+b')) as s:
        ...     s.get('https://i.sjtu.edu.cn')
        <Response [200 OK]>

    :param username: JAccount username.
    :param password: JAccount password.
    :param cookies: The cookie to be used on each request.
    :param ocr: The captcha :class:`Recognizer`.
    :param session_file: The file which a session is loaded from & saved to.
    :param retry: A list contains retry delays. If it's exhausted, an exception will be raised.
    :param proxies: The proxy to be used on each request.
    :param timeout: The timeout to be used on each request.
    :param base_url: Base url of backend APIs.
    :param _mocker_app: An WSGI application to send requests to (for debug or test purposes).
    """
    _client: httpx.Client  # httpx session
    _retry: list = [.5] * 5 + list(range(1, 5))  # retry list
    _ocr: Recognizer
    _username: str
    _password: str
    _session_file: Optional[FileTypes]
    _base_url: str

    def _secure_req(self, ref: Callable) -> Response:
        """
        Send a request using HTTPS explicitly to work around an upstream bug.

        :param ref: a partial request call.
        :return: the response of the original request.
        """
        try:
            return ref()
        except httpx.exceptions.NetworkError as e:
            req = e.request
            if not req.url.is_ssl:
                req.url = req.url.copy_with(scheme="https", port=None)
            else:
                raise e
            return self._client.send(req)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._client.close()
        if self._session_file:
            if isinstance(self._session_file, (io.RawIOBase, io.BufferedIOBase)):
                self._session_file.seek(0)
            self.dump(self._session_file)

    def __init__(self, username: str = "", password: str = "", cookies: CookieTypes = None, ocr: Recognizer = None,
                 session_file: FileTypes = None, retry: list = None, proxies: ProxiesTypes = None,
                 timeout: TimeoutTypes = None, base_url: str = "https://i.sjtu.edu.cn", _mocker_app=None):
        self._client = httpx.Client(app=_mocker_app, proxies=proxies)
        self._ocr = ocr if ocr else JCSSRecognizer()
        self._username = ""
        self._password = ""
        self._cache_store = {}
        self._base_url = base_url
        # noinspection PyTypeChecker
        self._session_file = None
        if retry:
            self._retry = retry
        if timeout:
            self.timeout = timeout

        if session_file:
            self.load(session_file)
            self._session_file = session_file

        if username and password:
            self.loads({"username": username, "password": password})
        elif cookies:
            self.loads({"cookies": cookies})

    def request(
            self,
            method: str,
            url: URLTypes,
            *,
            data: RequestData = None,
            files: RequestFiles = None,
            json: Any = None,
            params: QueryParamTypes = None,
            headers: HeaderTypes = None,
            cookies: CookieTypes = None,
            auth: AuthTypes = None,
            allow_redirects: bool = True,
            timeout: Union[TimeoutTypes, UnsetType] = UNSET,
            validate_session: bool = True,
            auto_renew: bool = True
    ) -> Response:
        """
        Send a request. If asked, validate the current session and renew it when necessary.

        :param method: HTTP method for the new `Request` object: `GET`, `OPTIONS`,
            `HEAD`, `POST`, `PUT`, `PATCH`, or `DELETE`.
        :param url: URL for the new `Request` object.
        :param params: (optional) Query parameters to include in the URL, as a
            string, dictionary, or list of two-tuples.
        :param data: (optional) Data to include in the body of the request, as a
            dictionary.
        :param files: (optional) A dictionary of upload files to include in the
            body of the request.
        :param json: (optional) A JSON serializable object to include in the body
            of the request.
        :param headers: (optional) Dictionary of HTTP headers to include in the
            request.
        :param cookies: (optional) Dictionary of Cookie items to include in the
            request.
        :param auth: (optional) An authentication class to use when sending the
            request.
        :param timeout: (optional) The timeout configuration to use when sending
            the request.
        :param allow_redirects: (optional) Enables or disables HTTP redirects.
        :param validate_session: (optional) Whether to validate the current session.
        :param auto_renew: (optional) Whether to renew the session when it expires. Works when validate_session is True.
        :return: an :class:`Response` object.
        """
        # complete the url in case base_url is omitted
        url_parsed = urlparse(url)
        url = url if all([getattr(url_parsed, attr) for attr in ["scheme", "netloc", "path"]]) \
            else urljoin(self._base_url, url)
        rtn = self._client.request(method=method,
                                   url=url,
                                   data=data,
                                   files=files,
                                   json=json,
                                   params=params,
                                   headers=headers,
                                   cookies=cookies,
                                   auth=auth,
                                   allow_redirects=allow_redirects,
                                   timeout=timeout)
        try:
            rtn.raise_for_status()
        except httpx.exceptions.HTTPError as e:
            if rtn.status_code == httpx.codes.SERVICE_UNAVAILABLE:
                raise ServiceUnavailable
            raise e
        if validate_session and rtn.url.full_path == "/xtgl/login_slogin.html":  # type: ignore
            if not auto_renew:
                raise SessionException("Session expired.")
            self._secure_req(partial(self.get, consts.LOGIN_URL, validate_session=False))  # refresh token
            # Sometimes JAccount OAuth token isn't expired
            if self.get(consts.HOME_URL,
                        validate_session=False).url.full_path == "/xtgl/login_slogin.html":  # type: ignore
                if self._username and self._password:
                    self.login(self._username, self._password)
                else:
                    raise SessionException("Session expired. Unable to renew session due to missing username or "
                                           "password")
            return self.request(method=method,
                                url=url,
                                data=data,
                                files=files,
                                json=json,
                                params=params,
                                headers=headers,
                                cookies=cookies,
                                auth=auth,
                                allow_redirects=allow_redirects,
                                timeout=timeout,
                                validate_session=validate_session,
                                auto_renew=False)  # disable auto_renew to avoid infinite recursion
        else:
            return rtn

    def get(
            self,
            url: URLTypes,
            *,
            params: QueryParamTypes = None,
            headers: HeaderTypes = None,
            cookies: CookieTypes = None,
            auth: AuthTypes = None,
            allow_redirects: bool = True,
            timeout: Union[TimeoutTypes, UnsetType] = UNSET,
            validate_session: bool = True,
            auto_renew: bool = True
    ) -> Response:
        """
        Send a GET request. If asked, validate the current session and renew it when necessary.

        :param url: URL for the new `Request` object.
        :param params: (optional) Query parameters to include in the URL, as a
            string, dictionary, or list of two-tuples.
        :param headers: (optional) Dictionary of HTTP headers to include in the
            request.
        :param cookies: (optional) Dictionary of Cookie items to include in the
            request.
        :param auth: (optional) An authentication class to use when sending the
            request.
        :param timeout: (optional) The timeout configuration to use when sending
            the request.
        :param allow_redirects: (optional) Enables or disables HTTP redirects.
        :param validate_session: (optional) Whether to validate the current session.
        :param auto_renew: (optional) Whether to renew the session when it expires. Works when validate_session is True.
        :return: an :class:`Response` object.
        """
        return self.request(
            "GET",
            url,
            params=params,
            headers=headers,
            cookies=cookies,
            auth=auth,
            allow_redirects=allow_redirects,
            timeout=timeout,
            validate_session=validate_session,
            auto_renew=auto_renew
        )

    def options(
            self,
            url: URLTypes,
            *,
            params: QueryParamTypes = None,
            headers: HeaderTypes = None,
            cookies: CookieTypes = None,
            auth: AuthTypes = None,
            allow_redirects: bool = True,
            timeout: Union[TimeoutTypes, UnsetType] = UNSET,
            validate_session: bool = True,
            auto_renew: bool = True
    ) -> Response:
        """
        Send a OPTIONS request. If asked, validate the current session and renew it when necessary.

        :param url: URL for the new `Request` object.
        :param params: (optional) Query parameters to include in the URL, as a
            string, dictionary, or list of two-tuples.
        :param headers: (optional) Dictionary of HTTP headers to include in the
            request.
        :param cookies: (optional) Dictionary of Cookie items to include in the
            request.
        :param auth: (optional) An authentication class to use when sending the
            request.
        :param timeout: (optional) The timeout configuration to use when sending
            the request.
        :param allow_redirects: (optional) Enables or disables HTTP redirects.
        :param validate_session: (optional) Whether to validate the current session.
        :param auto_renew: (optional) Whether to renew the session when it expires. Works when validate_session is True.
        :return: an :class:`Response` object.
        """
        return self.request(
            "OPTIONS",
            url,
            params=params,
            headers=headers,
            cookies=cookies,
            auth=auth,
            allow_redirects=allow_redirects,
            timeout=timeout,
            validate_session=validate_session,
            auto_renew=auto_renew
        )

    def head(
            self,
            url: URLTypes,
            *,
            params: QueryParamTypes = None,
            headers: HeaderTypes = None,
            cookies: CookieTypes = None,
            auth: AuthTypes = None,
            allow_redirects: bool = False,  # NOTE: Differs to usual default.
            timeout: Union[TimeoutTypes, UnsetType] = UNSET,
            validate_session: bool = True,
            auto_renew: bool = True
    ) -> Response:
        """
        Send a HEAD request. If asked, validate the current session and renew it when necessary.

        :param url: URL for the new `Request` object.
        :param params: (optional) Query parameters to include in the URL, as a
            string, dictionary, or list of two-tuples.
        :param headers: (optional) Dictionary of HTTP headers to include in the
            request.
        :param cookies: (optional) Dictionary of Cookie items to include in the
            request.
        :param auth: (optional) An authentication class to use when sending the
            request.
        :param timeout: (optional) The timeout configuration to use when sending
            the request.
        :param allow_redirects: (optional) Enables or disables HTTP redirects.
        :param validate_session: (optional) Whether to validate the current session.
        :param auto_renew: (optional) Whether to renew the session when it expires. Works when validate_session is True.
        :return: an :class:`Response` object.
        """
        return self.request(
            "HEAD",
            url,
            params=params,
            headers=headers,
            cookies=cookies,
            auth=auth,
            allow_redirects=allow_redirects,
            timeout=timeout,
            validate_session=validate_session,
            auto_renew=auto_renew
        )

    def post(
            self,
            url: URLTypes,
            *,
            data: RequestData = None,
            files: RequestFiles = None,
            json: Any = None,
            params: QueryParamTypes = None,
            headers: HeaderTypes = None,
            cookies: CookieTypes = None,
            auth: AuthTypes = None,
            allow_redirects: bool = True,
            timeout: Union[TimeoutTypes, UnsetType] = UNSET,
            validate_session: bool = True,
            auto_renew: bool = True
    ) -> Response:
        """
        Send a POST request. If asked, validate the current session and renew it when necessary.

        :param url: URL for the new `Request` object.
        :param params: (optional) Query parameters to include in the URL, as a
            string, dictionary, or list of two-tuples.
        :param data: (optional) Data to include in the body of the request, as a
            dictionary.
        :param files: (optional) A dictionary of upload files to include in the
            body of the request.
        :param json: (optional) A JSON serializable object to include in the body
            of the request.
        :param headers: (optional) Dictionary of HTTP headers to include in the
            request.
        :param cookies: (optional) Dictionary of Cookie items to include in the
            request.
        :param auth: (optional) An authentication class to use when sending the
            request.
        :param timeout: (optional) The timeout configuration to use when sending
            the request.
        :param allow_redirects: (optional) Enables or disables HTTP redirects.
        :param validate_session: (optional) Whether to validate the current session.
        :param auto_renew: (optional) Whether to renew the session when it expires. Works when validate_session is True.
        :return: an :class:`Response` object.
        """
        return self.request(
            "POST",
            url,
            data=data,
            files=files,
            json=json,
            params=params,
            headers=headers,
            cookies=cookies,
            auth=auth,
            allow_redirects=allow_redirects,
            timeout=timeout,
            validate_session=validate_session,
            auto_renew=auto_renew
        )

    def put(
            self,
            url: URLTypes,
            *,
            data: RequestData = None,
            files: RequestFiles = None,
            json: Any = None,
            params: QueryParamTypes = None,
            headers: HeaderTypes = None,
            cookies: CookieTypes = None,
            auth: AuthTypes = None,
            allow_redirects: bool = True,
            timeout: Union[TimeoutTypes, UnsetType] = UNSET,
            validate_session: bool = True,
            auto_renew: bool = True
    ) -> Response:
        """
        Send a PUT request. If asked, validate the current session and renew it when necessary.

        :param url: URL for the new `Request` object.
        :param params: (optional) Query parameters to include in the URL, as a
            string, dictionary, or list of two-tuples.
        :param data: (optional) Data to include in the body of the request, as a
            dictionary.
        :param files: (optional) A dictionary of upload files to include in the
            body of the request.
        :param json: (optional) A JSON serializable object to include in the body
            of the request.
        :param headers: (optional) Dictionary of HTTP headers to include in the
            request.
        :param cookies: (optional) Dictionary of Cookie items to include in the
            request.
        :param auth: (optional) An authentication class to use when sending the
            request.
        :param timeout: (optional) The timeout configuration to use when sending
            the request.
        :param allow_redirects: (optional) Enables or disables HTTP redirects.
        :param validate_session: (optional) Whether to validate the current session.
        :param auto_renew: (optional) Whether to renew the session when it expires. Works when validate_session is True.
        :return: an :class:`Response` object.
        """
        return self.request(
            "PUT",
            url,
            data=data,
            files=files,
            json=json,
            params=params,
            headers=headers,
            cookies=cookies,
            auth=auth,
            allow_redirects=allow_redirects,
            timeout=timeout,
            validate_session=validate_session,
            auto_renew=auto_renew
        )

    def patch(
            self,
            url: URLTypes,
            *,
            data: RequestData = None,
            files: RequestFiles = None,
            json: Any = None,
            params: QueryParamTypes = None,
            headers: HeaderTypes = None,
            cookies: CookieTypes = None,
            auth: AuthTypes = None,
            allow_redirects: bool = True,
            timeout: Union[TimeoutTypes, UnsetType] = UNSET,
            validate_session: bool = True,
            auto_renew: bool = True
    ) -> Response:
        """
        Sends a PATCH request. If asked, validates the current session and renews it when necessary.

        :param url: URL for the new `Request` object.
        :param params: (optional) Query parameters to include in the URL, as a
            string, dictionary, or list of two-tuples.
        :param data: (optional) Data to include in the body of the request, as a
            dictionary.
        :param files: (optional) A dictionary of upload files to include in the
            body of the request.
        :param json: (optional) A JSON serializable object to include in the body
            of the request.
        :param headers: (optional) Dictionary of HTTP headers to include in the
            request.
        :param cookies: (optional) Dictionary of Cookie items to include in the
            request.
        :param auth: (optional) An authentication class to use when sending the
            request.
        :param timeout: (optional) The timeout configuration to use when sending
            the request.
        :param allow_redirects: (optional) Enables or disables HTTP redirects.
        :param validate_session: (optional) Whether to validate the current session.
        :param auto_renew: (optional) Whether to renew the session when it expires. Works when validate_session is True.
        :return: an :class:`Response` object.
        """
        return self.request(
            "PATCH",
            url,
            data=data,
            files=files,
            json=json,
            params=params,
            headers=headers,
            cookies=cookies,
            auth=auth,
            allow_redirects=allow_redirects,
            timeout=timeout,
            validate_session=validate_session,
            auto_renew=auto_renew
        )

    def delete(
            self,
            url: URLTypes,
            *,
            params: QueryParamTypes = None,
            headers: HeaderTypes = None,
            cookies: CookieTypes = None,
            auth: AuthTypes = None,
            allow_redirects: bool = True,
            timeout: Union[TimeoutTypes, UnsetType] = UNSET,
            validate_session: bool = True,
            auto_renew: bool = True
    ) -> Response:
        """
        Sends a DELETE request. If asked, validates the current session and renews it when necessary.

        :param url: URL for the new `Request` object.
        :param params: (optional) Query parameters to include in the URL, as a
            string, dictionary, or list of two-tuples.
        :param headers: (optional) Dictionary of HTTP headers to include in the
            request.
        :param cookies: (optional) Dictionary of Cookie items to include in the
            request.
        :param auth: (optional) An authentication class to use when sending the
            request.
        :param timeout: (optional) The timeout configuration to use when sending
            the request.
        :param allow_redirects: (optional) Enables or disables HTTP redirects.
        :param validate_session: (optional) Whether to validate the current session.
        :param auto_renew: (optional) Whether to renew the session when it expires. Works when validate_session is True.
        :return: an :class:`Response` object.
        """
        return self.request(
            "DELETE",
            url,
            params=params,
            headers=headers,
            cookies=cookies,
            auth=auth,
            allow_redirects=allow_redirects,
            timeout=timeout,
            validate_session=validate_session,
            auto_renew=auto_renew
        )

    def login(self, username: str, password: str):
        """
        Log in JAccount using given username & password.

        :param username: JAccount username.
        :param password: JAccount password.
        :raises LoginException: Failed to login after several attempts.
        """
        self._cache_store = {}
        for i in self._retry:
            login_page_req = self._secure_req(partial(self.get, consts.LOGIN_URL, validate_session=False, headers=consts.HEADERS))
            uuid = re.findall(r"(?<=uuid\": ').*(?=')", login_page_req.text)[0]
            login_params = {k: v[0] for k, v in parse_qs(urlparse(str(login_page_req.url)).query).items()}

            captcha_img = self.get(consts.CAPTCHA_URL,
                                   params={"uuid": uuid, "t": int(time.time() * 1000)}).content
            captcha = self._ocr.recognize(captcha_img)

            login_params.update({"v": "", "uuid": uuid, "user": username, "pass": password, "captcha": captcha})
            result = self._secure_req(
                partial(self.post, consts.LOGIN_POST_URL, params=login_params, headers=consts.HEADERS))
            if "err=" not in result.url.query:  # type: ignore
                self._username = username
                self._password = password
                return

            time.sleep(i)

        raise LoginException

    def logout(self, purge_session: bool = True):
        """
        Log out JAccount.

        :param purge_session: (optional) Whether to purge local session info. May causes inconsistency, so use with
            caution.
        """
        cookie_bak = self._client.cookies
        self.get(consts.LOGOUT_URL, params={"t": int(time.time() * 1000), "login_type": ""}, validate_session=False)
        if purge_session:
            self._username = ''
            self._password = ''
        else:
            self._client.cookies = cookie_bak

    def loads(self, d: dict):
        """
        Read a session from a given dict. A warning will be given if username or password field is missing.

        :param d: a dict contains a session.
        """
        renew_required = True

        if "cookies" in d.keys() and d["cookies"]:
            if isinstance(d["cookies"], httpx.models.Cookies):
                cj = d["cookies"]  # type: ignore
            elif isinstance(d["cookies"], dict):
                cj = CookieJar()  # type: ignore
                # noinspection PyTypeHints
                cj._cookies = d["cookies"]  # type: ignore
            else:
                raise TypeError
            try:
                self.cookies = cj
                renew_required = False
            except SessionException:
                pass
        else:
            self._cookies = {}

        if "username" not in d.keys() or "password" not in d.keys() or not d["username"] or not d["password"]:
            warnings.warn("Missing username or password field", LoadWarning)
            self._username = ""
            self._password = ""
            renew_required = False
        else:
            self._username = d["username"]
            self._password = d["password"]

        if renew_required:
            self.login(self._username, self._password)

    def load(self, fp: FileTypes):
        """
        Read a session from a given file. A warning will be given if username or password field is missing.

        :param fp: a binary file object / filepath contains a session.
        """
        if isinstance(fp, (io.RawIOBase, io.BufferedIOBase)):
            try:
                conf = pickle.load(fp)
            except EOFError:
                conf = {}
        elif isinstance(fp, (str, Path)):
            try:
                with open(fp, mode="rb") as f:
                    conf = pickle.load(f)
            except EOFError:
                conf = {}
        else:
            raise TypeError
        self.loads(conf)

    # noinspection PyProtectedMember
    def dumps(self) -> dict:
        """
        Return a dict represents the current session. A warning will be given if username or password field is missing.

        :return: a dict represents the current session.
        """
        if not self._username or not self._password:
            warnings.warn("Missing username or password field", DumpWarning)
        return {"username": self._username, "password": self._password,
                "cookies": self._client.cookies.jar._cookies}  # type: ignore

    def dump(self, fp: FileTypes):
        """
        Write the current session to a given file. A warning will be given if username or password field is missing.

        :param fp: a binary file object/ filepath as the destination of session data.
        """
        if isinstance(fp, (io.RawIOBase, io.BufferedIOBase)):
            pickle.dump(self.dumps(), fp)
        elif isinstance(fp, (str, Path)):
            with open(fp, mode="wb") as f:
                pickle.dump(self.dumps(), f)
        else:
            raise TypeError

    @property
    def proxies(self) -> Dict[str, SyncDispatcher]:
        """ Get or set the proxy to be used on each request. """
        return self._client.proxies

    @property
    def _cookies(self) -> CookieTypes:
        """ Get or set the cookie to be used on each request. This protected property skips session validation. """
        return self._client.cookies

    @_cookies.setter
    def _cookies(self, new_cookie: CookieTypes):
        self._cache_store = {}
        # noinspection PyTypeHints
        self._client.cookies = new_cookie  # type: ignore

    @property
    def cookies(self) -> CookieTypes:
        """
        Get or set the cookie to be used on each request. Session validation is performed on each set event.

        :raises SessionException: when given cookie doesn't contain a valid session.
        """
        return self._client.cookies

    @cookies.setter
    def cookies(self, new_cookie: CookieTypes):
        bak_cookie = self._client.cookies
        # noinspection PyTypeHints
        self._client.cookies = new_cookie  # type: ignore
        self._secure_req(partial(self.get, consts.LOGIN_URL, validate_session=False, headers=consts.HEADERS))  # refresh JSESSION token
        if self.get(consts.HOME_URL, validate_session=False).url.full_path == "/xtgl/login_slogin.html":  # type: ignore
            self._client.cookies = bak_cookie
            raise SessionException("Invalid cookies. You may skip this validation by setting _cookies")
        self._cache_store = {}

    @property
    def timeout(self) -> TimeoutTypes:
        """ Get or set the timeout to be used on each request. """
        return self._client.timeout

    @timeout.setter
    def timeout(self, new_timeout: TimeoutTypes):
        if isinstance(new_timeout, (tuple, float, int, httpx.Timeout)) or new_timeout is None:
            self._client.timeout = httpx.Timeout(new_timeout)
        else:
            raise TypeError

    @property
    def base_url(self) -> str:
        """ Base url of backend APIs. """
        return self._base_url



import datetime
import json
from os import path

from marshmallow import ValidationError
import pytest

from pysjtu.models import CourseRange, LogicEnum, Ranking
from pysjtu.models.selection import Gender, LessonTime
from pysjtu.schemas import ExamSchema, GPAQueryParamsSchema, GPASchema, LibCourseSchema, ScheduleCourseSchema, \
    ScoreFactorSchema, ScoreSchema, SelectionClassSchema, SelectionCourseSchema, SelectionSectorSchema, \
    SelectionSharedInfoSchema
from pysjtu.schemas.base import StrBool
from pysjtu.schemas.schedule import CreditHourDetail


@pytest.fixture()
def resp_loader():
    RESP_DIR = path.join(path.dirname(path.abspath(__file__)), 'resources/resp')

    def _resp_loader(name):
        with open(path.join(RESP_DIR, name + ".json"), encoding="utf-8") as f:
            return json.load(f)

    return _resp_loader


def test_selection_course_schema(resp_loader):
    raw_resp = resp_loader("selection_course")
    schema = SelectionCourseSchema()
    course = schema.load(raw_resp)

    assert course == {
        "name": "问题求解与实践",
        "credit": 3.0,
        "course_id": "CS241",
        "internal_course_id": "_CS241",
        "class_name": "(2020-2021-1)-CS241-1",
        "class_id": "A86B96D4FB8A3CFEE055F8163ED16360",
        "students_registered": 11
    }


def test_selection_class_schema(resp_loader):
    raw_resp = resp_loader("selection_class")
    schema = SelectionClassSchema()
    _class = schema.load(raw_resp)

    assert _class == {
        "class_id": "A86B96D4FB8A3CFEE055F8163ED16360",
        "register_id": "2914000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000",
        "teachers": [("陈雨亭", "副教授"), ("沈艳艳", "讲师(高校)")],
        "locations": ["中院312", "中院312", "中院312", "中院312"],
        "time": [
            LessonTime(2, [range(1, 13)], [range(9, 11)]),
            LessonTime(2, [range(1, 13)], [range(9, 11)]),
            LessonTime(5, [range(1, 13, 2)], [range(3, 5)]),
            LessonTime(5, [range(1, 13)], [range(3, 5), range(7, 9)])
        ],
        "course_type": ["必修", "测试"],
        "remark": None,
        "students_planned": 80
    }

    class_2 = schema.load({"xkbz": "N/A", **raw_resp})
    assert class_2["remark"] == "N/A"


def test_selection_sector_schema(resp_loader):
    raw_resp = resp_loader("selection_sector")
    schema = SelectionSectorSchema()
    sector = schema.load(raw_resp)

    assert sector.task_type == 1
    assert sector.xkly == 1
    assert sector.pe_op_param == 30
    assert sector.sector_type_id == "0"
    assert sector.txbsfrl == 1
    assert sector.kkbk == 0
    assert sector.include_other_grades is False
    assert sector.include_other_majors is False
    assert sector.sfznkx is False
    assert sector.zdkxms == 0

    sector.course_type_code = "01"
    sector.xkkz_id = "A000000000000B76E055F8163ED16360"
    assert {k: str(v) for k, v in schema.dump(sector).items()} == \
           {"kklxdm": "01", "xkkz_id": "A000000000000B76E055F8163ED16360",
            **raw_resp}


def test_selection_shared_info_schema(resp_loader):
    raw_resp = resp_loader("selection_shared_info")
    schema = SelectionSharedInfoSchema()
    shared_info = schema.load(raw_resp)

    assert shared_info.term == "02"
    assert shared_info.selection_year == 2020
    assert shared_info.selection_term == 3
    assert shared_info.major_id == "000000"
    assert shared_info.student_grade == 2019
    assert shared_info.natural_class_id == "F1900001"
    assert shared_info.self_selecting_status == 1
    assert shared_info.ccdm == "0"
    assert shared_info.student_type_code == 999
    assert shared_info.gender == Gender.female
    assert shared_info.field_id == "wfx"
    assert shared_info.student_background == 99999

    assert {k: str(v) for k, v in schema.dump(shared_info).items()} == raw_resp


def test_schedule_credit_hour_detail_field():
    field = CreditHourDetail()
    assert field.deserialize("task_1:3.0,task_2:0.5") == {"task_1": 3.0, "task_2": 0.5}
    assert field.deserialize("-") == {"N/A": 0}


def test_str_bool_field():
    field = StrBool()
    assert field.deserialize("0") is False
    assert field.deserialize("1") is True
    with pytest.raises(ValidationError):
        field.deserialize("-1")
    assert field.serialize("test", {"test": False}) == "0"
    assert field.serialize("test", {"test": True}) == "1"
    with pytest.raises(ValidationError):
        field.serialize("test", {"test": -1})


def test_schedule_course_schema(resp_loader):
    raw_resp = resp_loader("schedule_course")
    schema = ScheduleCourseSchema()
    schedule_course = schema.load(raw_resp)

    assert schedule_course.name == "军事理论"
    assert schedule_course.day == 1 and isinstance(schedule_course.day, int)
    assert schedule_course.week == [6, 10, range(14, 16)]
    assert schedule_course.time == range(1, 3)
    assert schedule_course.location == "下院306"
    assert schedule_course.credit == 1.0 and isinstance(schedule_course.credit, float)
    assert schedule_course.assessment == "考试"
    assert schedule_course.remark == "无"
    assert schedule_course.teacher_name == ["闫成"]
    assert schedule_course.teacher_title == ["讲师(高校)"]
    assert schedule_course.course_id == "TH004"
    assert schedule_course.class_name == "(2019-2020-1)-TH004-2"
    assert schedule_course.class_id == "90BEF7AA5D657072E0530200A8C06959"
    assert schedule_course.hour_total == 16 and isinstance(schedule_course.hour_total, int)
    assert schedule_course.hour_remark == {"理论": 24, "其他": 24}
    assert schedule_course.hour_week == 1 and isinstance(schedule_course.hour_week, int)
    assert schedule_course.field == "无方向"


def test_score_factor_schema(resp_loader):
    raw_resp = resp_loader("score_factor")
    schema = ScoreFactorSchema(many=True)
    score_factors = schema.load(raw_resp)

    assert score_factors[0].name == "平时"
    assert score_factors[0].percentage == 0.4
    assert score_factors[0].score == "93"

    assert score_factors[1].name == "期末"
    assert score_factors[1].percentage == 0.6
    assert score_factors[1].score == "B+"


def test_score_schema(resp_loader):
    raw_resp = resp_loader("score")
    schema = ScoreSchema()
    score = schema.load(raw_resp)

    assert score.name == "大学化学"
    assert score.teacher == "麦亦勇"
    assert score.score == "91"
    assert score.credit == 2 and isinstance(score.gp, float)
    assert score.gp == 4.0 and isinstance(score.gp, float)
    assert not score.invalid and isinstance(score.invalid, bool)
    assert score.course_type == "主修"
    assert score.category == "个性化教育课程"
    assert score.score_type == "正常考试"
    assert score.method == "考试"
    assert score.course_id == "CA001"
    assert score.class_name == "(2019-2020-1)-CA001-10"
    assert score.class_id == "90DD3334A9410650E0530200A8C03235"


def test_exam_schema(resp_loader):
    raw_resp = resp_loader("exam")
    schema = ExamSchema()
    exam = schema.load(raw_resp)

    assert exam.name == "2019-2020-1数学期中考"
    assert exam.location == "东上院509"
    assert exam.seat == 1
    assert exam.course_id == "MA248"
    assert exam.course_name == "高等数学I"
    assert exam.class_name == "(2019-2020-1)-MA248-7"
    assert not exam.rebuild and isinstance(exam.rebuild, bool)
    assert exam.credit == 6.0 and isinstance(exam.credit, float)
    assert not exam.self_study and isinstance(exam.self_study, bool)
    assert exam.date == datetime.date(2019, 11, 6)
    assert exam.time == [datetime.time(13, 10), datetime.time(15, 10)]


def test_gpa_query_params_schema_load(resp_loader):
    raw_resp = resp_loader("gpa_query_params")
    schema = GPAQueryParamsSchema()
    gpa_query_params = schema.load(raw_resp)

    assert gpa_query_params.gp_round == 9
    assert gpa_query_params.gpa_round == 9
    assert gpa_query_params.exclude_gp == "缓考"
    assert gpa_query_params.exclude_gpa == "缓考"
    assert gpa_query_params.course_whole == ["TH020", "TH009"]
    assert gpa_query_params.has_roll and isinstance(gpa_query_params.has_roll, bool)
    assert gpa_query_params.registered is None
    assert gpa_query_params.attending and isinstance(gpa_query_params.attending, bool)


def test_gpa_query_params_schema_dump(resp_loader):
    raw_resp = resp_loader("gpa_query_params")
    schema = GPAQueryParamsSchema()
    gpa_query_params = schema.load(raw_resp)

    gpa_query_params.end_term = 2019

    gpa_query_params.condition_logic = 0
    with pytest.raises(TypeError):
        schema.dump(gpa_query_params)
    gpa_query_params.condition_logic = LogicEnum.AND

    gpa_query_params.makeup_as_60 = True
    gpa_query_params.rebuild_as_60 = True

    gpa_query_params.course_range = 0
    with pytest.raises(TypeError):
        schema.dump(gpa_query_params)
    gpa_query_params.course_range = CourseRange.ALL

    gpa_query_params.ranking = 0
    with pytest.raises(TypeError):
        schema.dump(gpa_query_params)
    gpa_query_params.ranking = Ranking.GRADE_AND_FIELD

    gpa_query_params.has_roll = False
    gpa_query_params.registered = True
    gpa_query_params.attending = None

    dump_dict = schema.dump(gpa_query_params)

    assert dump_dict == {'zczt': 1, 'bjjd': '缓考', 'xjzt': 0, 'bjpjf': '缓考', 'qsXnxq': '', 'tjfw': 'njzy',
                         'kch_ids': 'TH020,TH009', 'sspjfblws': 9, 'tjgx': 0, 'pjjdblws': 9, 'zzXnxq': 2019,
                         'kcfw': 'qbkc', 'alsfj': 'bkcx'}


def test_gpa_schema(resp_loader):
    raw_resp = resp_loader("gpa")
    schema = GPASchema()
    gpa = schema.load(raw_resp)

    assert gpa.total_score == 999
    assert gpa.course_count == 14
    assert gpa.fail_count == 0
    assert gpa.total_credit == 32.5
    assert gpa.acquired_credit == 32.5
    assert gpa.failed_credit == 0
    assert gpa.pass_rate == 1
    assert gpa.gp == 98.987654321
    assert gpa.gp_ranking == 21
    assert gpa.gpa == 4.123456789
    assert gpa.gpa_ranking == 20
    assert gpa.total_students == 99


def test_lib_course_schema_1(resp_loader):
    schema = LibCourseSchema()
    raw_resp = resp_loader("lib_course_1")
    lib_course = schema.load(raw_resp)

    assert lib_course.name == "高等数学I"
    assert lib_course.day == 1 and isinstance(lib_course.day, int)
    assert lib_course.week == [range(1, 17)]
    assert lib_course.time == range(1, 3)
    assert lib_course.location == "东下院415"
    assert lib_course.locations == ["东下院415", "东下院415", "东下院415"]
    assert lib_course.faculty == "数学科学学院"
    assert lib_course.credit == 6.0 and isinstance(lib_course.credit, float)
    assert lib_course.teacher == ["王铭"]
    assert lib_course.course_id == "MA248"
    assert lib_course.class_name == "(2019-2020-1)-MA248-10"
    assert lib_course.class_id == "908698F1AEF44587E0530200A8C0E301"
    assert lib_course.class_composition == ["F1902122", "F1902123", "F1902127", "F1902142"]
    assert lib_course.hour_total == 96 and isinstance(lib_course.hour_total, int)
    assert lib_course.hour_remark == {"理论": 6.0}
    assert lib_course.seats == 168
    assert lib_course.students_elected == 100
    assert lib_course.students_planned == 115


def test_lib_course_schema_2(resp_loader):
    schema = LibCourseSchema()
    raw_resp = resp_loader("lib_course_2")
    lib_course = schema.load(raw_resp)

    assert lib_course.name == "高等数学A1"
    assert lib_course.day == 1 and isinstance(lib_course.day, int)
    assert lib_course.week == [range(1, 14, 2)]
    assert lib_course.time == range(1, 3)
    assert lib_course.location == "东下院102"
    assert lib_course.locations == ["东下院102", "东上院101", "上院100"]
    assert lib_course.faculty == "密西根学院"
    assert lib_course.credit == 4.0 and isinstance(lib_course.credit, float)
    assert lib_course.teacher == ["OLGA DANILKINA"]
    assert lib_course.course_id == "VV156"
    assert lib_course.class_name == "(2019-2020-1)-VV156-1"
    assert lib_course.class_id == "895A258003C03F72E0530200A8C0E75F"
    assert lib_course.class_composition == ["2019电子与计算机工程", "2019机械类"]
    assert lib_course.hour_total == 56 and isinstance(lib_course.hour_total, int)
    assert lib_course.hour_remark == {"理论": 4.0, "实验": 1.0}
    assert lib_course.seats == 126
    assert lib_course.students_elected == 113
    assert lib_course.students_planned == 300
